#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "05/09/2019"

import datetime
import os
import json
import time
import requests

from suds.client import Client
from suds.sax.date import DateTime
from suds.transport.https import HttpAuthenticated

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


def getDataFromURL(url):
    if "http_proxy" in os.environ:
        os.environ["http_proxy"] = ""
    response = requests.get(url)
    data = {"statusCode": response.status_code}
    if response.status_code == 200:
        data["data"] = json.loads(response.text)[0]
    else:
        data["text"] = response.text
    return data


def getRawDataFromURL(url):
    if "http_proxy" in os.environ:
        os.environ["http_proxy"] = ""
    response = requests.get(url)
    data = {"statusCode": response.status_code}
    if response.status_code == 200:
        data["content"] = response.content
    else:
        data["text"] = response.text
    return data


def getWdslRoot():
    dictConfig = UtilsConfig.getTaskConfig("ISPyB")
    wdslRoot = dictConfig["ispyb_ws_url"]
    return wdslRoot


def getTransport():
    transport = None
    logger = UtilsLogging.getLogger()
    if "ISPyB_user" not in os.environ:
        logger.error("No ISPyB user name defined as environment variable!")
    elif "ISPyB_pass" not in os.environ:
        logger.error("No ISPyB password defined as environment variable!")
    else:
        ispybUserName = os.environ["ISPyB_user"]
        ispybPassword = os.environ["ISPyB_pass"]
        transport = HttpAuthenticated(username=ispybUserName, password=ispybPassword)
    return transport


def getCollectionWebService():
    logger = UtilsLogging.getLogger()
    collectionWdsl = getToolsForCollectionWebService()
    transport = getTransport()
    if transport is None:
        logger.error(
            "No transport defined, ISPyB web service client cannot be instantiated."
        )
        collectionWSClient = None
    else:
        collectionWSClient = Client(collectionWdsl, transport=transport, cache=None)
    return collectionWSClient


def getToolsForCollectionWebService():
    return os.path.join(getWdslRoot(), "ispybWS", "ToolsForCollectionWebService?wsdl")


def getToolsForAutoprocessingWebService():
    return os.path.join(
        getWdslRoot(), "ispybWS", "ToolsForAutoprocessingWebService?wsdl"
    )


def getAutoprocessingWebService():
    logger = UtilsLogging.getLogger()
    collectionWdsl = getToolsForAutoprocessingWebService()
    transport = getTransport()
    if transport is None:
        logger.error(
            "No transport defined, ISPyB web service client cannot be instantiated."
        )
        autoprocessingWSClient = None
    else:
        autoprocessingWSClient = Client(collectionWdsl, transport=transport, cache=None)
    return autoprocessingWSClient


def findDataCollection(dataCollectionId, client=None):
    e = None
    dataCollectionWS3VO = None
    noTrials = 5
    logger = UtilsLogging.getLogger()
    try:
        if client is None:
            client = getCollectionWebService()
        if client is None:
            logger.error(
                "No web service client available, cannot contact findDataCollection web service."
            )
        elif dataCollectionId is None:
            logger.error(
                "No dataCollectionId given, cannot contact findDataCollection web service."
            )
        else:
            dataCollectionWS3VO = client.service.findDataCollection(dataCollectionId)
    except Exception as e:
        logger.error(
            "ISPyB error for findDataCollection: {0}, {1} trials left".format(
                e, noTrials
            )
        )
    return dataCollectionWS3VO


def findDataCollectionFromFileLocationAndFileName(imagePath, client=None):
    logger = UtilsLogging.getLogger()
    dataCollectionWS3VO = None
    noTrials = 10
    fileLocation = os.path.dirname(imagePath)
    fileName = os.path.basename(imagePath)
    if fileName.endswith(".h5"):
        prefix = UtilsImage.getPrefix(fileName)
        imageNumber = UtilsImage.getImageNumber(fileName)
        fileName = "{0}_{1:04d}.h5".format(prefix, imageNumber)
    try:
        if client is None:
            client = getCollectionWebService()
        if client is None:
            logger.error(
                "No web service client available, cannot contact findDataCollectionFromFileLocationAndFileName web service."
            )
        elif fileLocation is None:
            logger.error(
                "No fileLocation given, cannot contact findDataCollectionFromFileLocationAndFileName web service."
            )
        elif fileName is None:
            logger.error(
                "No fileName given, cannot contact findDataCollectionFromFileLocationAndFileName web service."
            )
        else:
            dataCollectionWS3VO = (
                client.service.findDataCollectionFromFileLocationAndFileName(
                    fileLocation, fileName
                )
            )
    except Exception as e:
        logger.error(
            "ISPyB error for findDataCollectionFromFileLocationAndFileName: {0}, {1} trials left".format(
                e, noTrials
            )
        )
        raise e
    if dataCollectionWS3VO is None:
        time.sleep(1)
        if noTrials == 0:
            logger.error("No data collections found for path {0}".format(imagePath))
        else:
            logger.warning(
                "Cannot find {0} in ISPyB - retrying, {1} trials left".format(
                    imagePath, noTrials
                )
            )
    return dataCollectionWS3VO


def setImageQualityIndicatorsPlot(dataCollectionId, plotFile, csvFile):
    logger = UtilsLogging.getLogger()
    client = getCollectionWebService()
    if client is None:
        logger.error(
            "No web service client available, cannot contact setImageQualityIndicatorsPlot web service."
        )
    returnDataCollectionId = client.service.setImageQualityIndicatorsPlot(
        dataCollectionId, plotFile, csvFile
    )
    return returnDataCollectionId


def storeOrUpdateAutoProcProgram(
    programs,
    commandline,
    status,
    message=None,
    start_time=None,
    end_time=None,
    environment=None,
    record_time_stamp=None,
    autoProcProgramId=None,
):
    if start_time is None:
        start_time = datetime.datetime.now()
    if end_time is None:
        end_time = datetime.datetime.now()
    if record_time_stamp is None:
        record_time_stamp = datetime.datetime.now()
    client = getAutoprocessingWebService()
    autoProcProgramId = client.service.storeOrUpdateAutoProcProgram(
        arg0=autoProcProgramId,
        processingCommandLine=commandline,
        processingPrograms=programs,
        processingStatus=status,
        processingMessage=message,
        processingStartTime=DateTime(start_time),
        processingEndTime=DateTime(end_time),
        processingEnvironment=environment,
        recordTimeStamp=record_time_stamp,
    )
    return autoProcProgramId


def storeOrUpdateAutoProcProgramAttachment(
    auto_proc_program_id,
    file_type,
    file_path,
    record_time_stamp=None,
    auto_proc_program_attachment_id=None,
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_program_attachment_id = (
        client.service.storeOrUpdateAutoProcProgramAttachment(
            arg0=auto_proc_program_attachment_id,
            fileType=file_type,
            fileName=os.path.basename(file_path),
            filePath=os.path.dirname(file_path),
            recordTimeStamp=record_time_stamp,
            autoProcProgramId=auto_proc_program_id,
        )
    )
    return auto_proc_program_attachment_id


def storeOrUpdateAutoProcIntegration(
    auto_proc_program_id,
    dataCollection_id,
    start_image_number=None,
    end_image_number=None,
    auto_proc_integration_id=None,
    refined_detector_distance=None,
    refined_x_beam=None,
    refined_y_beam=None,
    rotation_axis_x=None,
    rotation_axis_y=None,
    rotation_axis_z=None,
    beam_vector_x=None,
    beam_vector_y=None,
    beam_vector_z=None,
    cell_a=None,
    cell_b=None,
    cell_c=None,
    cell_alpha=None,
    cell_beta=None,
    cell_gamma=None,
    record_time_stamp=None,
    anomalous=None,
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_integration_id = client.service.storeOrUpdateAutoProcIntegration(
        arg0=auto_proc_integration_id,
        autoProcProgramId=auto_proc_program_id,
        startImageNumber=start_image_number,
        endImageNumber=end_image_number,
        refinedDetectorDistance=refined_detector_distance,
        refinedXbeam=refined_x_beam,
        refinedYbeam=refined_y_beam,
        rotationAxisX=rotation_axis_x,
        rotationAxisY=rotation_axis_y,
        rotationAxisZ=rotation_axis_z,
        beamVectorX=beam_vector_x,
        beamVectorY=beam_vector_y,
        beamVectorZ=beam_vector_z,
        cellA=cell_a,
        cellB=cell_b,
        cellC=cell_c,
        cellAlpha=cell_alpha,
        cellBeta=cell_beta,
        cellGamma=cell_gamma,
        recordTimeStamp=record_time_stamp,
        anomalous=anomalous,
        dataCollectionId=dataCollection_id,
    )
    return auto_proc_integration_id


def storeOrUpdateAutoProc(
    auto_proc_program_id,
    space_group,
    refined_cell_a,
    refined_cell_b,
    refined_cell_c,
    refined_cell_alpha,
    refined_cell_beta,
    refined_cell_gamma,
    auto_proc_id=None,
    record_time_stamp=None,
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_id = client.service.storeOrUpdateAutoProc(
        arg0=auto_proc_id,
        autoProcProgramId=auto_proc_program_id,
        spaceGroup=space_group,
        refinedCellA=refined_cell_a,
        refinedCellB=refined_cell_b,
        refinedCellC=refined_cell_c,
        refinedCellAlpha=refined_cell_alpha,
        refinedCellBeta=refined_cell_beta,
        refinedCellGamma=refined_cell_gamma,
        recordTimeStamp=record_time_stamp,
    )
    return auto_proc_id


def storeOrUpdateAutoProcScaling(
    auto_proc_id,
    resolution_ellipsoid_axis_11=None,
    resolution_ellipsoid_axis_12=None,
    resolution_ellipsoid_axis_13=None,
    resolution_ellipsoid_axis_21=None,
    resolution_ellipsoid_axis_22=None,
    resolution_ellipsoid_axis_23=None,
    resolution_ellipsoid_axis_31=None,
    resolution_ellipsoid_axis_32=None,
    resolution_ellipsoid_axis_33=None,
    resolution_ellipsoid_value_1=None,
    resolution_ellipsoid_value_2=None,
    resolution_ellipsoid_value_3=None,
    auto_proc_scaling_id=None,
    record_time_stamp=None,
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_scaling_id = client.service.storeOrUpdateAutoProcScaling(
        arg0=auto_proc_scaling_id,
        autoProcId=auto_proc_id,
        recordTimeStamp=record_time_stamp,
        resolutionEllipsoidAxis11=resolution_ellipsoid_axis_11,
        resolutionEllipsoidAxis12=resolution_ellipsoid_axis_12,
        resolutionEllipsoidAxis13=resolution_ellipsoid_axis_13,
        resolutionEllipsoidAxis21=resolution_ellipsoid_axis_21,
        resolutionEllipsoidAxis22=resolution_ellipsoid_axis_22,
        resolutionEllipsoidAxis23=resolution_ellipsoid_axis_23,
        resolutionEllipsoidAxis31=resolution_ellipsoid_axis_31,
        resolutionEllipsoidAxis32=resolution_ellipsoid_axis_32,
        resolutionEllipsoidAxis33=resolution_ellipsoid_axis_33,
        resolutionEllipsoidValue1=resolution_ellipsoid_value_1,
        resolutionEllipsoidValue2=resolution_ellipsoid_value_2,
        resolutionEllipsoidValue3=resolution_ellipsoid_value_3,
    )
    return auto_proc_scaling_id


def storeOrUpdateAutoProcScalingHasInt(
    auto_proc_integration_id,
    auto_proc_scaling_id,
    record_time_stamp=None
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_scaling_has_int_id = client.service.storeOrUpdateAutoProcScalingHasInt(
        arg0=None,
        autoProcIntegrationId=auto_proc_integration_id,
        autoProcScalingId=auto_proc_scaling_id,
        recordTimeStamp=record_time_stamp
    )
    return auto_proc_scaling_has_int_id


def storeOrUpdateAutoProcScalingStatistics(
    scaling_statistics_type,
    resolution_limit_low,
    resolution_limit_high,
    r_merge,
    r_meas_within_i_plus_i_minus,
    r_meas_all_i_plus_i_minus,
    r_pim_within_i_plus_i_minus,
    r_pim_all_i_plus_i_minus,
    ntotal_observations,
    ntotal_unique_observations,
    mean_iover_sig_i,
    completeness,
    multiplicity,
    anomalous_completeness,
    anomalous_multiplicity,
    auto_proc_scaling_id,
    cc_half,
    cc_ano,
    comments=None,
    anomalous=None,
    sig_ano=None,
    isa=None,
    completeness_spherical=None,
    anomalous_completeness_spherical=None,
    completeness_ellipsoidal=None,
    anomalous_completeness_ellipsoidal=None,
    fractional_partial_bias=None,
    auto_proc_scaling_statistics_id=None,
    record_time_stamp=None,
):
    if record_time_stamp is None:
        record_time_stamp = DateTime(datetime.datetime.now())
    client = getAutoprocessingWebService()
    auto_proc_scaling_statistics_id = (
        client.service.storeOrUpdateAutoProcScalingStatistics(
            arg0=auto_proc_scaling_statistics_id,
            scalingStatisticsType=scaling_statistics_type,
            comments=comments,
            resolutionLimitLow=resolution_limit_low,
            resolutionLimitHigh=resolution_limit_high,
            rmerge=r_merge,
            rmeasWithinIplusIminus=r_meas_within_i_plus_i_minus,
            rmeasAllIplusIminus=r_meas_all_i_plus_i_minus,
            rpimWithinIplusIminus=r_pim_within_i_plus_i_minus,
            rpimAllIplusIminus=r_pim_all_i_plus_i_minus,
            fractionalPartialBias=fractional_partial_bias,
            nTotalObservations=ntotal_observations,
            nTotalUniqueObservations=ntotal_unique_observations,
            meanIoverSigI=mean_iover_sig_i,
            completeness=completeness,
            multiplicity=multiplicity,
            anomalousCompleteness=anomalous_completeness,
            anomalousMultiplicity=anomalous_multiplicity,
            recordTimeStamp=record_time_stamp,
            anomalous=anomalous,
            autoProcScalingId=auto_proc_scaling_id,
            ccHalf=cc_half,
            ccAno=cc_ano,
            sigAno=sig_ano,
            isa=isa,
            completenessSpherical=completeness_spherical,
            anomalousCompletenessSpherical=anomalous_completeness_spherical,
            completenessEllipsoidal=completeness_ellipsoidal,
            anomalousCompletenessEllipsoidal=anomalous_completeness_ellipsoidal,
        )
    )
    return auto_proc_scaling_statistics_id



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from builtins import RuntimeError

import numpy
import shlex
import distro
import shutil
import base64
import pathlib
import matplotlib.pyplot as plt

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader
from edna2.tasks.ISPyBTasks import ISPyBRetrieveDataCollection

from edna2.utils import UtilsPath
from edna2.utils import UtilsImage
from edna2.utils import UtilsIspyb
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging
from edna2.utils import UtilsDetector

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginDozor-v1.0/plugins/EDPluginDozorv1_0.py
# mxv1/plugins/EDPluginControlDozor-v1.0/plugins/EDPluginControlDozorv1_0.py

# matplotlib.use('Agg')

logger = UtilsLogging.getLogger()

# Default values for ESRF Pilatus2M : ID30a1: 1,776; 826,894
IX_MIN_PILATUS_2M = 1
IX_MAX_PILATUS_2M = 776
IY_MIN_PILATUS_2M = 826
IY_MAX_PILATUS_2M = 894

# Default values for ESRF Pilatus6M : id23eh1: 1,1230; 1228, 1298
IX_MIN_PILATUS_6M = 1
IX_MAX_PILATUS_6M = 1230
IY_MIN_PILATUS_6M = 1228
IY_MAX_PILATUS_6M = 1298

# Default values for ESRF Eiger4M : ID30a3: 1,1120; 1025,1140
IX_MIN_EIGER_4M = 1
IX_MAX_EIGER_4M = 1120
IY_MIN_EIGER_4M = 1025
IY_MAX_EIGER_4M = 1140

# Default parameters

DEFAULT_FRACTION_POLARIZATION = 0.99
DEFAULT_IMAGE_STEP = 1
MAX_BATCH_SIZE = 5000


class ExecDozor(AbstractTask):  # pylint: disable=too-many-instance-attributes
    """
    The ExecDozor is responsible for executing the 'dozor' program.
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "required": [
                "detectorType",
                "exposureTime",
                "spotSize",
                "detectorDistance",
                "wavelength",
                "orgx",
                "orgy",
                "oscillationRange",
                "firstImageNumber",
                "numberImages",
                "nameTemplateImage",
            ],
            "properties": {
                "detectorType": {"type": "string"},
                "beamline": {"type": "string"},
                "exposureTime": {"type": "number"},
                "spotSize": {"type": "integer"},
                "spotLevel": {"type": "integer"},
                "detectorDistance": {"type": "number"},
                "wavelength": {"type": "number"},
                "fractionPolarization": {"type": "number"},
                "orgx": {"type": "number"},
                "orgy": {"type": "number"},
                "oscillationRange": {"type": "number"},
                "imageStep": {"type": "number"},
                "startingAngle": {"type": "number"},
                "firstImageNumber": {"type": "integer"},
                "numberImages": {"type": "integer"},
                "nameTemplateImage": {"type": "string"},
                "wedgeNumber": {"type": "integer"},
                "radiationDamage": {"type": "boolean"},
                "overlap": {"type": "number"},
                "doDozorM": {"type": "boolean"},
                "doSubmit": {"type": "boolean"}
            },
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "required": ["imageDozor"],
            "properties": {
                "imageDozor": {
                    "type": "array",
                    "items": {"$ref": self.getSchemaUrl("imageDozor.json")},
                },
                "halfDoseTime": {"type": "number"},
                "dozorPlot": {"type": "string"},
                "plotmtvFile": {"type": "string"},
                "pngPlots": {"type": "array", "items": {"type": "string"}},
                "dozorAllFile": {"type": "string"},
            },
        }

    def run(self, inData):
        doSubmit = inData.get("doSubmit", False)
        doDozorM = inData.get("doDozorM", False)
        commands = self.generateCommands(inData)
        with open(str(self.getWorkingDirectory() / "dozor.dat"), "w") as f:
            f.write(commands)
        # Create dozor command line
        if doSubmit:
            executable = UtilsConfig.get(self, "slurm_executable", "dozor")
            partition = UtilsConfig.get(self, "slurm_partition", None)
            noCores = UtilsConfig.get(self, "noCores", 10)
        else:
            executable = UtilsConfig.get(self, "executable", "dozor")
            partition = None
            noCores = None
        commandLine = executable + " -pall"
        if doDozorM:
            commandLine += " -mesh"
        if "radiationDamage" in inData:
            commandLine += " -rd dozor.dat"
        else:
            commandLine += " -p dozor.dat"
        self.setLogFileName("dozor.log")
        self.runCommandLine(
            commandLine,
            doSubmit=doSubmit,
            partition=partition,
            noCores=noCores
        )
        log = self.getLog()
        outData = self.parseOutput(
            inData, log, doDozorM=doDozorM, workingDir=self.getWorkingDirectory()
        )
        return outData

    def generateCommands(self, inData):
        """
        This method creates the input file for dozor
        """
        ixMin = None
        ixMax = None
        iyMin = None
        iyMax = None
        detectorType = inData["detectorType"]
        nx = UtilsDetector.getNx(detectorType)
        ny = UtilsDetector.getNy(detectorType)
        pixelSize = UtilsDetector.getPixelsize(detectorType)
        sitePrefix = UtilsConfig.get(self, "site_prefix")
        doSubmit = inData.get("doSubmit", False)
        if (
            sitePrefix is not None
            and "beamline" in inData
            and inData["beamline"] is not None
        ):
            # Try to read corresponding config file
            site = sitePrefix + inData["beamline"]
            taskConfig = UtilsConfig.getTaskConfig(self.__class__.__name__, site)
            ixMin = int(taskConfig["ix_min"])
            ixMax = int(taskConfig["ix_max"])
            iyMin = int(taskConfig["iy_min"])
            iyMax = int(taskConfig["iy_max"])
        elif detectorType == "pilatus2m":
            ixMin = IX_MIN_PILATUS_2M
            ixMax = IX_MAX_PILATUS_2M
            iyMin = IY_MIN_PILATUS_2M
            iyMax = IY_MAX_PILATUS_2M
        elif detectorType == "pilatus6m":
            ixMin = IX_MIN_PILATUS_6M
            ixMax = IX_MAX_PILATUS_6M
            iyMin = IY_MIN_PILATUS_6M
            iyMax = IY_MAX_PILATUS_6M
        elif detectorType == "eiger4m":
            ixMin = IX_MIN_EIGER_4M
            ixMax = IX_MAX_EIGER_4M
            iyMin = IY_MIN_EIGER_4M
            iyMax = IY_MAX_EIGER_4M
        if inData["nameTemplateImage"].endswith("h5"):
            library = self.getLibrary("hdf5", doSubmit=doSubmit)
        else:
            library = self.getLibrary("cbf", doSubmit=doSubmit)
        processInfo = "name template: {0}".format(
            os.path.basename(inData["nameTemplateImage"])
        )
        processInfo += ", first image no: {0}".format(inData["firstImageNumber"])
        processInfo += ", no images: {0}".format(inData["numberImages"])
        command = "!\n"
        command += "detector %s\n" % detectorType
        command += "library %s\n" % library
        command += "nx %d\n" % nx
        command += "ny %d\n" % ny
        command += "pixel %f\n" % pixelSize
        command += "exposure %.3f\n" % inData["exposureTime"]
        command += "spot_size %d\n" % inData["spotSize"]
        command += "spot_level %d\n" % inData.get("spotLevel", 6)
        command += "detector_distance %.3f\n" % inData["detectorDistance"]
        command += "X-ray_wavelength %.3f\n" % inData["wavelength"]
        fractionPolarization = inData.get(
            "fractionPolarization", DEFAULT_FRACTION_POLARIZATION
        )
        command += "fraction_polarization %.3f\n" % fractionPolarization
        command += "pixel_min 0\n"
        command += "pixel_max 64000\n"
        if ixMin is not None:
            command += "ix_min %d\n" % ixMin
            command += "ix_max %d\n" % ixMax
            command += "iy_min %d\n" % iyMin
            command += "iy_max %d\n" % iyMax
        badZona = UtilsConfig.get(self, "bad_zona", None)
        if badZona is not None:
            command += "bad_zona %s\n" % badZona
        command += "orgx %.1f\n" % inData["orgx"]
        command += "orgy %.1f\n" % inData["orgy"]
        command += "oscillation_range %.3f\n" % inData["oscillationRange"]
        imageStep = inData.get("imageStep", DEFAULT_IMAGE_STEP)
        command += "image_step %.3f\n" % imageStep
        first_image_number = inData["firstImageNumber"]
        overall_starting_angle = (
            inData["startingAngle"]
            - (first_image_number - 1) * inData["oscillationRange"]
        )
        command += "starting_angle %.3f\n" % overall_starting_angle
        command += "first_image_number %d\n" % first_image_number
        command += "number_images %d\n" % inData["numberImages"]
        if "wedgeNumber" in inData:
            command += "wedge_number %d\n" % inData["wedgeNumber"]
        command += "name_template_image %s\n" % inData["nameTemplateImage"]
        command += "end\n"
        # logger.debug('command: {0}'.format(command))
        return command

    def parseOutput(self, inData, output, doDozorM=False, workingDir=None):
        """
        This method parses the output of dozor
        """
        resultDozor = {
            "imageDozor": []  # list of dict. each dict contains spotFile and Image_path
        }
        # Create template for image name
        template = inData["nameTemplateImage"]
        # HDF5 workaround
        if template.endswith(".h5"):
            template = template.replace("1_??????", "??????")
        noWildCards = template.count("?")
        template = template.replace("?" * noWildCards, "{0:0" + str(noWildCards) + "}")
        # Skip the four first lines
        listOutput = output.split("\n")[6:]

        for line in listOutput:
            # Remove '|'
            listLine = shlex.split(line.replace("|", " "))
            if len(listLine) > 0 and listLine[0].isdigit():
                imageDozor = {}
                imageNumber = int(listLine[0])
                overlap = inData.get("overlap", 0.0)
                angle = (
                    inData["startingAngle"]
                    + (imageNumber - inData["firstImageNumber"])
                    * (inData["oscillationRange"] - overlap)
                    + inData["oscillationRange"] / 2.0
                )
                imageDozor["number"] = imageNumber
                imageDozor["image"] = template.format(imageNumber)
                imageDozor["angle"] = angle
                imageDozor["spotsNumOf"] = None
                imageDozor["spotsIntAver"] = None
                imageDozor["spotsResolution"] = None
                imageDozor["mainScore"] = None
                imageDozor["spotScore"] = None
                imageDozor["visibleResolution"] = 40
                try:
                    if listLine[5].startswith("-") or len(listLine) < 11:
                        imageDozor["spotsNumOf"] = int(listLine[1])
                        imageDozor["spotsIntAver"] = self.parseDouble(listLine[2])
                        imageDozor["spotsRFactor"] = self.parseDouble(listLine[3])
                        imageDozor["spotsResolution"] = self.parseDouble(listLine[4])
                        imageDozor["mainScore"] = self.parseDouble(listLine[8])
                        imageDozor["spotScore"] = self.parseDouble(listLine[9])
                        imageDozor["visibleResolution"] = self.parseDouble(listLine[10])
                    else:
                        imageDozor["spotsNumOf"] = int(listLine[1])
                        imageDozor["spotsIntAver"] = self.parseDouble(listLine[2])
                        imageDozor["spotsRfactor"] = self.parseDouble(listLine[3])
                        imageDozor["spotsResolution"] = self.parseDouble(listLine[4])
                        imageDozor["powderWilsonScale"] = self.parseDouble(listLine[5])
                        imageDozor["powderWilsonBfactor"] = self.parseDouble(
                            listLine[6]
                        )
                        imageDozor["powderWilsonResolution"] = self.parseDouble(
                            listLine[7]
                        )
                        imageDozor["powderWilsonCorrelation"] = self.parseDouble(
                            listLine[8]
                        )
                        imageDozor["powderWilsonRfactor"] = self.parseDouble(
                            listLine[9]
                        )
                        imageDozor["mainScore"] = self.parseDouble(listLine[10])
                        imageDozor["spotScore"] = self.parseDouble(listLine[11])
                        imageDozor["visibleResolution"] = self.parseDouble(listLine[12])
                except Exception as e:
                    logger.warning("Exception caught when parsing Dozor log!")
                    logger.warning(e)
                # ExecDozor spot file
                if workingDir is not None:
                    spotFile = os.path.join(
                        str(workingDir), "%05d.spot" % imageDozor["number"]
                    )
                    if os.path.exists(spotFile):
                        imageDozor["spotFile"] = spotFile
                #                #print imageDozor['marshal()
                resultDozor["imageDozor"].append(imageDozor)
            elif line.startswith("h"):
                resultDozor["halfDoseTime"] = line.split("=")[1].split()[0]
        # Check if mtv plot file exists
        if workingDir is not None:
            mtvFileName = "dozor_rd.mtv"
            mtvFilePath = os.path.join(str(workingDir), mtvFileName)
            if os.path.exists(mtvFilePath):
                resultDozor["plotmtvFile"] = mtvFilePath
                resultDozor["pngPlots"] = self.generatePngPlots(mtvFilePath, workingDir)
        if doDozorM:
            # Create 'dozor_all_result' file
            dozorAllFile = str(self.getWorkingDirectory() / "dozor_all")
            os.system(
                "cat {0}/*.all > {1}".format(
                    str(self.getWorkingDirectory()), dozorAllFile
                )
            )
            resultDozor["dozorAllFile"] = dozorAllFile
        return resultDozor

    @classmethod
    def parseDouble(cls, value):
        returnValue = None
        try:
            returnValue = float(value)
        except Exception as ex:
            errorMessage = "Error when trying to parse " " + value + " ": %r" % ex
            logger.error(errorMessage)
        return returnValue

    @classmethod
    def generatePngPlots(cls, plotmtvFile, workingDir):
        listXSFile = []
        # Create plot dictionary
        with open(str(plotmtvFile)) as f:
            listLines = f.readlines()
        listPlots = []
        dictPlotList = []
        dictSubPlot = {}
        index = 0
        while index < len(listLines):
            # print('0' + listLines[index])
            if listLines[index].startswith("$"):
                dictPlot = {}
                dictPlotList = []
                listPlots.append(dictPlot)
                dictPlot["plotList"] = dictPlotList
                index += 1
                dictPlot["name"] = listLines[index].split("'")[1]
                index += 1
                while listLines[index].startswith("%"):
                    listLine = listLines[index].split("=")
                    label = listLine[0][1:].strip()
                    if "'" in listLine[1]:
                        value = listLine[1].split("'")[1]
                    else:
                        value = listLine[1]
                    value = value.replace("\n", "").strip()
                    dictPlot[label] = value
                    index += 1
            elif listLines[index].startswith("#"):
                dictSubPlot = {}
                dictPlotList.append(dictSubPlot)
                plotName = listLines[index].split("#")[1]
                dictSubPlot["name"] = plotName.replace("\n", "").strip()
                index += 1
                while listLines[index].startswith("%"):
                    listLine = listLines[index].split("=")
                    label = listLine[0][1:].strip()
                    if "'" in listLine[1]:
                        value = listLine[1].split("'")[1]
                    else:
                        value = listLine[1]
                    value = value.replace("\n", "").strip()
                    dictSubPlot[label] = value
                    index += 1
                dictSubPlot["xValues"] = []
                dictSubPlot["yValues"] = []
            else:
                listData = listLines[index].replace("\n", "").split()
                dictSubPlot["xValues"].append(float(listData[0]))
                dictSubPlot["yValues"].append(float(listData[1]))
                index += 1
        # Generate the plots
        for mtvplot in listPlots:
            listLegend = []
            xmin = None
            xmax = None
            ymin = None
            ymax = None
            for subPlot in mtvplot["plotList"]:
                xmin1 = min(subPlot["xValues"])
                if xmin is None or xmin > xmin1:
                    xmin = xmin1
                xmax1 = max(subPlot["xValues"])
                if xmax is None or xmax < xmax1:
                    xmax = xmax1
                ymin1 = min(subPlot["yValues"])
                if ymin is None or ymin > ymin1:
                    ymin = ymin1
                ymax1 = max(subPlot["yValues"])
                if ymax is None or ymax < ymax1:
                    ymax = ymax1
            if "xmin" in mtvplot:
                xmin = float(mtvplot["xmin"])
            if "ymin" in mtvplot:
                ymin = float(mtvplot["ymin"])
            plt.xlim(xmin, xmax)
            plt.ylim(ymin, ymax)
            plt.xlabel(mtvplot["xlabel"])
            plt.ylabel(mtvplot["ylabel"])
            plt.title(mtvplot["name"])
            for subPlot in mtvplot["plotList"]:
                if "markercolor" in subPlot:
                    style = "bs-."
                else:
                    style = "r"
                plt.plot(subPlot["xValues"], subPlot["yValues"], style, linewidth=2)
                listLegend.append(subPlot["linelabel"])
            plt.legend(listLegend, loc="lower right")
            mtvPlotName = mtvplot["name"].replace(" ", "").replace(".", "_")
            plotPath = os.path.join(str(workingDir), mtvPlotName + ".png")
            plt.savefig(plotPath, bbox_inches="tight", dpi=75)
            plt.close()
            listXSFile.append(plotPath)
        return listXSFile

    def getLibrary(self, libraryType, doSubmit=False):
        libraryName = "library_" + libraryType
        if doSubmit:
            libraryName += "_ubuntu_20.04"
        else:
            idName, version, codename = distro.linux_distribution()
            if "Debian" in idName:
                libraryName += "_debian_"
            elif idName == "Ubuntu":
                libraryName += "_ubuntu_"
            else:
                raise RuntimeError("ExecDozor: unknown os name {0}".format(idName))
            libraryName += version
        library = UtilsConfig.get(self, libraryName)
        if library is None:
            raise RuntimeError(
                "ExecDozor: library configuration {0} not found".format(libraryName)
            )
        return library


class ControlDozor(AbstractTask):
    def __init__(self, inData, workingDirectorySuffix=None):
        AbstractTask.__init__(
            self, inData, workingDirectorySuffix=workingDirectorySuffix
        )
        self.hasOverlap = False
        self.overlap = 0.0

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "dataCollectionId": {
                    "anyOf": [
                        {"type": "integer"},
                        {"type": "null"}
                    ]
                },
                "processDirectory": {"type": "string"},
                "image": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "directory": {"type": "string"},
                "beamline": {"type": "string"},
                "template": {"type": "string"},
                "startNo": {"type": "integer"},
                "endNo": {"type": "integer"},
                "batchSize": {"type": "integer"},
                "hdf5BatchSize": {"type": "integer"},
                "wedgeNumber": {"type": "integer"},
                "radiationDamage": {"type": "string"},
                "keepCbfTmpDirectory": {"type": "boolean"},
                "doISPyBUpload": {"type": "boolean"},
                "doDozorM": {"type": "boolean"},
                "returnSpotList": {"type": "boolean"},
            },
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "required": ["imageQualityIndicators", "detectorType"],
            "properties": {
                "imageQualityIndicators": {
                    "type": "array",
                    "items": {"$ref": self.getSchemaUrl("imageQualityIndicators.json")},
                },
                "detectorType": {"type": "string"},
                "halfDoseTime": {"type": "number"},
                "dozorPlot": {"type": "string"},
                "pathToCbfDirectory": {"type": "string"},
                "pngPlots": {"type": "string"},
                "dozorAllFile": {"type": "string"},
            },
        }

    def run(self, inData):
        # workingDirectory = self.getWorkingDirectory()
        outData = {}
        controlDozorAllFile = None
        listDozorAllFile = []
        hasHdf5Prefix = False
        detectorType = None
        returnSpotList = inData.get("returnSpotList", False)
        # Check doDozorM
        doDozorM = inData.get("doDozorM", False)
        # Check if connection to ISPyB needed
        batchSize, dictImage = self.determineBatchsize(inData)
        # Check overlap
        overlap = inData.get("overlap", self.overlap)
        if overlap != 0:
            self.hasOverlap = True
        logger.debug("ExecDozor batch size: {0}".format(batchSize))
        listAllBatches = self.createListOfBatches(
            dictImage.keys(), batchSize, self.hasOverlap
        )
        outData["imageQualityIndicators"] = []
        for listBatch in listAllBatches:
            outDataDozor, detectorType = self.runDozorTask(
                inData=inData,
                dictImage=dictImage,
                listBatch=listBatch,
                overlap=overlap,
                workingDirectory=str(self.getWorkingDirectory()),
                hasHdf5Prefix=hasHdf5Prefix,
                hasOverlap=self.hasOverlap,
            )
            if outDataDozor is not None:
                for imageDozor in outDataDozor["imageDozor"]:
                    imageQualityIndicators = {
                        "angle": imageDozor["angle"],
                        "number": imageDozor["number"],
                        "image": imageDozor["image"],
                        "dozorScore": imageDozor["mainScore"],
                        "dozorSpotScore": imageDozor["spotScore"],
                        "dozorSpotsNumOf": imageDozor["spotsNumOf"],
                        "dozorSpotsIntAver": imageDozor["spotsIntAver"],
                        "dozorSpotsResolution": imageDozor["spotsResolution"],
                        "dozorVisibleResolution": imageDozor["visibleResolution"],
                    }
                    if "spotFile" in imageDozor:
                        if os.path.exists(imageDozor["spotFile"]):
                            spotFile = imageDozor["spotFile"]
                            imageQualityIndicators["dozorSpotFile"] = spotFile
                            if returnSpotList:
                                numpyArray = numpy.loadtxt(spotFile, skiprows=3)
                                imageQualityIndicators[
                                    "dozorSpotList"
                                ] = base64.b64encode(numpyArray.tostring()).decode(
                                    "utf-8"
                                )
                                imageQualityIndicators["dozorSpotListShape"] = list(
                                    numpyArray.shape
                                )
                    outData["imageQualityIndicators"].append(imageQualityIndicators)
                if doDozorM:
                    listDozorAllFile.append(outDataDozor["dozorAllFile"])
        # Assemble all dozorAllFiles into one
        if doDozorM:
            controlDozorAllFile = str(self.getWorkingDirectory() / "dozor_all")
            os.system("touch {0}".format(controlDozorAllFile))
            for dozorAllFile in listDozorAllFile:
                command = "cat " + dozorAllFile + " >> " + controlDozorAllFile
                os.system(command)
        # Make plot if we have a data collection id
        if "dataCollectionId" in inData and inData["dataCollectionId"] is not None:
            if "processDirectory" in inData:
                processDirectory = pathlib.Path(inData["processDirectory"])
            else:
                processDirectory = self.getWorkingDirectory()
            dozorPlotPath, dozorCsvPath = self.makePlot(
                inData["dataCollectionId"], outData, self.getWorkingDirectory()
            )
            doIspybUpload = inData.get("doIspybUpload", False)
            if doIspybUpload:
                self.storeDataOnPyarch(
                    inData["dataCollectionId"],
                    dozorPlotPath=dozorPlotPath,
                    dozorCsvPath=dozorCsvPath,
                    workingDirectory=processDirectory,
                )
        #            xsDataResultControlDozor.halfDoseTime = edPluginDozor.dataOutput.halfDoseTime
        #            xsDataResultControlDozor.pngPlots = edPluginDozor.dataOutput.pngPlots
        #
        #            self.sendResultToMXCuBE(imageDozorBatchList)
        #            self.sendMessageToMXCuBE("Batch processed")
        #        if self.cbfTempDir is not None:
        #            if self.dataInput.keepCbfTmpDirectory is not None and self.dataInput.keepCbfTmpDirectory.value:
        #                self.dataOutput.pathToCbfDirectory = XSDataFile(XSDataString(self.cbfTempDir))
        #            else:
        #                shutil.rmtree(self.cbfTempDir)

        # Read the header from the first image in the batch
        outData["detectorType"] = detectorType
        if doDozorM:
            outData["dozorAllFile"] = controlDozorAllFile
        return outData

    def determineBatchsize(self, inData):
        dataCollectionId = inData.get("dataCollectionId", None)
        if dataCollectionId is not None:
            ispybInData = {"dataCollectionId": dataCollectionId}
            ispybTask = ISPyBRetrieveDataCollection(inData=ispybInData)
            ispybTask.execute()
            dataCollection = ispybTask.outData
            batchSize = UtilsConfig.get("ControlDozor", "batchSize")
            if batchSize is None:
                batchSize = dataCollection["numberOfImages"]
            if batchSize > MAX_BATCH_SIZE:
                batchSize = MAX_BATCH_SIZE
            if "overlap" in dataCollection and abs(dataCollection["overlap"]) > 1:
                self.hasOverlap = True
                self.overlap = dataCollection["overlap"]
            else:
                self.overlap = 0.0
            dictImage = self.createImageDictFromISPyB(dataCollection)
        else:
            # No connection to ISPyB, take parameters from input
            if "batchSize" in inData:
                batchSize = inData["batchSize"]
            else:
                batchSize = MAX_BATCH_SIZE
            dictImage = self.createImageDict(inData)
        return batchSize, dictImage

    @classmethod
    def runDozorTask(
        cls,
        inData,
        dictImage,
        listBatch,
        overlap,
        workingDirectory,
        hasHdf5Prefix,
        hasOverlap,
    ):
        doSubmit = inData.get("doSubmit", False)
        doDozorM = inData.get("doDozorM", False)
        outDataDozor = None
        image = dictImage[listBatch[0]]
        prefix = UtilsImage.getPrefix(image)
        suffix = UtilsImage.getSuffix(image)
        imageNumber = UtilsImage.getImageNumber(image)
        imageNumberOrig = imageNumber
        workingDirectorySuffixDozor = "{0:04d}_{1:04d}".format(
            imageNumber, imageNumber + len(listBatch) - 1
        )
        if image.endswith("h5"):
            hasHdf5Prefix = True
            prefix = UtilsImage.getPrefix(image)
            h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(
                image, hasOverlap=hasOverlap, isFastMesh=True
            )
            workingDirectorySuffix = "{0}_{1}_master_{2}".format(
                prefix, h5FileNumber, imageNumber
            )
            if hasOverlap:
                imageNumber = 1
        else:
            workingDirectorySuffix = "{0}_{1:04d}".format(
                prefix, UtilsImage.getImageNumber(image)
            )
        inDataReadHeader = {
            "imagePath": [image],
            "skipNumberOfImages": True,
            "hasOverlap": hasOverlap,
            "isFastMesh": True,
        }
        controlHeader = ReadImageHeader(
            inData=inDataReadHeader, workingDirectorySuffix=workingDirectorySuffix
        )
        controlHeader.execute()
        outDataHeader = controlHeader.outData
        subWedge = outDataHeader["subWedge"][0]
        experimentalCondition = subWedge["experimentalCondition"]
        beam = experimentalCondition["beam"]
        detector = experimentalCondition["detector"]
        detectorType = detector["type"]
        goniostat = experimentalCondition["goniostat"]
        inDataDozor = {
            "detectorType": detector["type"],
            "exposureTime": beam["exposureTime"],
            "detectorDistance": detector["distance"],
            "spotSize": 3,
            "wavelength": beam["wavelength"],
            "orgx": detector["beamPositionX"] / detector["pixelSizeX"],
            "orgy": detector["beamPositionY"] / detector["pixelSizeY"],
            "oscillationRange": goniostat["oscillationWidth"],
            "startingAngle": goniostat["rotationAxisStart"],
            "firstImageNumber": imageNumber,
            "numberImages": len(listBatch),
            "workingDirectory": workingDirectory,
            "overlap": overlap,
            "doSubmit": doSubmit,
            "doDozorM": doDozorM,
        }
        if "beamline" in inData:
            inDataDozor["beamline"] = inData["beamline"]
        if UtilsConfig.isEMBL():
            template = "{0}_?????.{1}".format(prefix, suffix)
        elif hasHdf5Prefix:
            template = "{0}_{1}_??????.{2}".format(prefix, h5FileNumber, suffix)
        elif imageNumber < 10000:
            template = "{0}_????.{1}".format(prefix, suffix)
        else:
            template = "{0}_?????.{1}".format(prefix, suffix)
        inDataDozor["nameTemplateImage"] = os.path.join(
            os.path.dirname(subWedge["image"][0]["path"]), template
        )
        if "wedgeNumber" in inData:
            inDataDozor["wedgeNumber"] = inData["wedgeNumber"]
        if "radiationDamage" in inData:
            inDataDozor["radiationDamage"] = inData["radiationDamage"]
        dozor = ExecDozor(
            inData=inDataDozor, workingDirectorySuffix=workingDirectorySuffixDozor
        )
        dozor.execute()
        if not dozor.isFailure():
            outDataDozor = dozor.outData
            if len(outDataDozor["imageDozor"]) == 1 and outDataDozor["imageDozor"][0][
                "image"
            ].endswith(".h5"):
                if outDataDozor["imageDozor"][0]["number"] != imageNumberOrig:
                    outDataDozor["imageDozor"][0]["number"] = imageNumberOrig
        return outDataDozor, detectorType

    def createGnuPlotFile(self, workingDirectory, csvFileName, outDataImageDozor):
        with open(str(workingDirectory / csvFileName), "w") as gnuplotFile:
            gnuplotFile.write("# Data directory: {0}\n".format(self.directory))
            gnuplotFile.write(
                "# File template: {0}\n".format(self.template.replace("%04d", "####"))
            )
            gnuplotFile.write(
                "# {0:>9s}{1:>16s}{2:>16s}{3:>16s}{4:>16s}{5:>16s}\n".format(
                    "'Image no'",
                    "'Angle'",
                    "'No of spots'",
                    "'Main score (*10)'",
                    "'Spot score'",
                    "'Visible res.'",
                )
            )
            for imageQualityIndicators in outDataImageDozor["imageQualityIndicators"]:
                gnuplotFile.write(
                    "{0:10d},{1:15.3f},{2:15d},{3:15.3f},{4:15.3f},{5:15.3f}\n".format(
                        imageQualityIndicators["number"],
                        imageQualityIndicators["angle"],
                        imageQualityIndicators["dozorSpotsNumOf"],
                        10 * imageQualityIndicators["dozorScore"],
                        imageQualityIndicators["dozorSpotScore"],
                        imageQualityIndicators["dozorVisibleResolution"],
                    )
                )

    def determineMinMaxParameters(self, outDataImageDozor):
        minImageNumber = None
        maxImageNumber = None
        minAngle = None
        maxAngle = None
        minDozorValue = None
        maxDozorValue = None
        minResolution = None
        maxResolution = None
        for imageQualityIndicators in outDataImageDozor["imageQualityIndicators"]:
            if (
                minImageNumber is None
                or minImageNumber > imageQualityIndicators["number"]
            ):
                minImageNumber = imageQualityIndicators["number"]
                minAngle = imageQualityIndicators["angle"]
            if (
                maxImageNumber is None
                or maxImageNumber < imageQualityIndicators["number"]
            ):
                maxImageNumber = imageQualityIndicators["number"]
                maxAngle = imageQualityIndicators["angle"]
            if (
                minDozorValue is None
                or minDozorValue > imageQualityIndicators["dozorScore"]
            ):
                minDozorValue = imageQualityIndicators["dozorScore"]
            if (
                maxDozorValue is None
                or maxDozorValue < imageQualityIndicators["dozorScore"]
            ):
                maxDozorValue = imageQualityIndicators["dozorScore"]

            # Min resolution: the higher the value the lower the resolution
            if (
                minResolution is None
                or minResolution < imageQualityIndicators["dozorVisibleResolution"]
            ):
                # Disregard resolution worse than 10.0
                if imageQualityIndicators["dozorVisibleResolution"] < 10.0:
                    minResolution = imageQualityIndicators["dozorVisibleResolution"]
            # Max resolution: the lower the number the better the resolution
            if (
                maxResolution is None
                or maxResolution > imageQualityIndicators["dozorVisibleResolution"]
            ):
                maxResolution = imageQualityIndicators["dozorVisibleResolution"]
        plotDict = {
            "minImageNumber": minImageNumber,
            "maxImageNumber": maxImageNumber,
            "minAngle": minAngle,
            "maxAngle": maxAngle,
            "minDozorValue": minDozorValue,
            "maxDozorValue": maxDozorValue,
            "minResolution": minResolution,
            "maxResolution": maxResolution,
        }
        return plotDict

    def determinePlotParameters(self, plotDict):
        xtics = ""
        minImageNumber = plotDict["minImageNumber"]
        maxImageNumber = plotDict["maxImageNumber"]
        minAngle = plotDict["minAngle"]
        maxAngle = plotDict["maxAngle"]
        minResolution = plotDict["minResolution"]
        maxResolution = plotDict["maxResolution"]
        minDozorValue = plotDict["minDozorValue"]
        maxDozorValue = plotDict["maxDozorValue"]
        if minImageNumber is not None and minImageNumber == maxImageNumber:
            minAngle -= 1.0
            maxAngle += 1.0
        noImages = maxImageNumber - minImageNumber + 1
        if noImages <= 4:
            minImageNumber -= 0.1
            maxImageNumber += 0.1
            deltaAngle = maxAngle - minAngle
            minAngle -= deltaAngle * 0.1 / noImages
            maxAngle += deltaAngle * 0.1 / noImages
            xtics = "1"
        if maxResolution is None or maxResolution > 0.8:
            maxResolution = 0.8
        else:
            maxResolution = int(maxResolution * 10.0) / 10.0
        if minResolution is None or minResolution < 4.5:
            minResolution = 4.5
        else:
            minResolution = int(minResolution * 10.0) / 10.0 + 1
        if maxDozorValue < 0.001 and minDozorValue < 0.001:
            yscale = "set yrange [-0.5:0.5]\n    set ytics 1"
        else:
            yscale = "set autoscale  y"
        plotDict = {
            "xtics": xtics,
            "yscale": yscale,
            "minImageNumber": minImageNumber,
            "maxImageNumber": maxImageNumber,
            "minAngle": minAngle,
            "maxAngle": maxAngle,
            "minDozorValue": minDozorValue,
            "maxDozorValue": maxDozorValue,
            "minResolution": minResolution,
            "maxResolution": maxResolution,
        }
        return plotDict

    def makePlot(self, dataCollectionId, outDataImageDozor, workingDirectory):
        plotFileName = "dozor_{0}.png".format(dataCollectionId)
        csvFileName = "dozor_{0}.csv".format(dataCollectionId)
        self.createGnuPlotFile(workingDirectory, csvFileName, outDataImageDozor)
        plotDict = self.determineMinMaxParameters(outDataImageDozor)
        plotDict = self.determinePlotParameters(plotDict)
        gnuplotScript = """#
set terminal png
set output '{dozorPlotFileName}'
set title '{title}'
set grid x2 y2
set xlabel 'Image number'
set x2label 'Angle (degrees)'
set y2label 'Resolution (A)'
set ylabel 'Number of spots / ExecDozor score (*10)'
set xtics {xtics} nomirror
set x2tics
set ytics nomirror
set y2tics
set xrange [{minImageNumber}:{maxImageNumber}]
set x2range [{minAngle}:{maxAngle}]
{yscale}
set y2range [{minResolution}:{maxResolution}]
set key below
plot '{dozorCsvFileName}' using 1:3 title 'Number of spots' axes x1y1 with points linetype rgb 'goldenrod' pointtype 7 pointsize 1.5, \
    '{dozorCsvFileName}' using 1:4 title 'ExecDozor score' axes x1y1 with points linetype 3 pointtype 7 pointsize 1.5, \
    '{dozorCsvFileName}' using 1:6 title 'Visible resolution' axes x1y2 with points linetype 1 pointtype 7 pointsize 1.5
""".format(
            title=self.template.replace("%04d", "####"),
            dozorPlotFileName=plotFileName,
            dozorCsvFileName=csvFileName,
            minImageNumber=plotDict["minImageNumber"],
            maxImageNumber=plotDict["maxImageNumber"],
            minAngle=plotDict["minAngle"],
            maxAngle=plotDict["maxAngle"],
            minResolution=plotDict["minResolution"],
            maxResolution=plotDict["maxResolution"],
            xtics=plotDict["xtics"],
            yscale=plotDict["yscale"],
        )
        pathGnuplotScript = str(workingDirectory / "gnuplot.sh")
        with open(pathGnuplotScript, "w") as f:
            f.write(gnuplotScript)
        oldCwd = os.getcwd()
        os.chdir(str(workingDirectory))
        gnuplot = UtilsConfig.get(self, "gnuplot", "gnuplot")
        os.system("{0} {1}".format(gnuplot, pathGnuplotScript))
        os.chdir(oldCwd)
        dozorPlotPath = workingDirectory / plotFileName
        dozorCsvPath = workingDirectory / csvFileName
        return dozorPlotPath, dozorCsvPath

    @classmethod
    def storeDataOnPyarch(
        cls, dataCollectionId, dozorPlotPath, dozorCsvPath, workingDirectory
    ):
        resultsDirectory = pathlib.Path(workingDirectory) / "results"
        try:
            if not resultsDirectory.exists():
                resultsDirectory.mkdir(parents=True, mode=0o755)
            dozorPlotResultPath = resultsDirectory / dozorPlotPath.name
            dozorCsvResultPath = resultsDirectory / dozorCsvPath.name
            shutil.copy(dozorPlotPath, dozorPlotResultPath)
            shutil.copy(dozorCsvPath, dozorCsvResultPath)
        except Exception as e:
            logger.warning(
                "Couldn't copy files to results directory: {0}".format(resultsDirectory)
            )
            logger.warning(e)
        try:
            # Create paths on pyarch
            dozorPlotPyarchPath = UtilsPath.createPyarchFilePath(dozorPlotResultPath)
            dozorCsvPyarchPath = UtilsPath.createPyarchFilePath(dozorCsvResultPath)
            if not os.path.exists(os.path.dirname(dozorPlotPyarchPath)):
                os.makedirs(os.path.dirname(dozorPlotPyarchPath), 0o755)
            shutil.copy(dozorPlotResultPath, dozorPlotPyarchPath)
            shutil.copy(dozorCsvResultPath, dozorCsvPyarchPath)
            # Upload to data collection
            dataCollectionId = UtilsIspyb.setImageQualityIndicatorsPlot(
                dataCollectionId, dozorPlotPyarchPath, dozorCsvPyarchPath
            )
        except Exception as e:
            logger.warning("Couldn't copy files to pyarch.")
            logger.warning(e)

    def createImageDict(self, inData):
        # Create dictionary of all images with the image number as key
        dictImage = {}
        if "image" in inData and len(inData["image"]) > 0:
            listImage = inData["image"]
            pathToFirstImage = listImage[0]
            self.directory = os.path.dirname(pathToFirstImage)
            self.template = os.path.basename(pathToFirstImage).replace("0001", "####")
        else:
            # Create list of images
            listImage = []
            self.directory = inData["directory"]
            if "#" in inData["template"]:
                noWildCards = inData["template"].count("#")
                self.template = inData["template"].replace(
                    "#" * noWildCards, "%0" + str(noWildCards) + "d"
                )
            else:
                self.template = inData["template"]
            startNo = int(inData["startNo"])
            endNo = int(inData["endNo"])
            for imageIndex in range(startNo, endNo + 1):
                imageName = self.template % imageIndex
                imagePath = os.path.join(self.directory, imageName)
                listImage.append(imagePath)
        for imagePath in listImage:
            imageNo = UtilsImage.getImageNumber(imagePath)
            dictImage[imageNo] = imagePath
        return dictImage

    def createImageDictFromISPyB(self, dataCollection):
        # Create dictionary of all images with the image number as key
        dictImage = {}
        # Create list of images
        listImage = []
        self.directory = dataCollection["imageDirectory"]
        self.template = dataCollection["fileTemplate"]
        startNo = dataCollection["startImageNumber"]
        endNo = (
            dataCollection["startImageNumber"] + dataCollection["numberOfImages"] - 1
        )
        for imageIndex in range(startNo, endNo + 1):
            imageName = self.template % imageIndex
            imagePath = os.path.join(self.directory, imageName)
            listImage.append(imagePath)
        for image in listImage:
            imageNo = UtilsImage.getImageNumber(image)
            dictImage[imageNo] = image
        return dictImage

    @classmethod
    def createListOfBatches(cls, listImage, batchSize, overlap=False):
        # Create the list of batches containing the image no
        listAllBatches = []
        listImagesInBatch = []
        indexBatch = 0
        indexNextImageInBatch = None
        listImageSorted = sorted(listImage)
        for imageNo in listImageSorted:
            if indexNextImageInBatch is None:
                # This image can be appended to this batch
                indexBatch = 1
                indexNextImageInBatch = imageNo + 1
                listImagesInBatch.append(imageNo)
                if batchSize == 1:
                    listAllBatches.append(listImagesInBatch)
                    listImagesInBatch = []
                    indexNextImageInBatch = None
            elif imageNo != indexNextImageInBatch or indexBatch == batchSize:
                # A new batch must be started
                indexBatch = 1
                listAllBatches.append(listImagesInBatch)
                listImagesInBatch = [imageNo]
                indexNextImageInBatch = imageNo + 1
            else:
                # This image can be appended to this batch
                listImagesInBatch.append(imageNo)
                indexNextImageInBatch += 1
                indexBatch += 1
        if listImagesInBatch:
            listAllBatches.append(listImagesInBatch)
        if overlap:
            # Split up batches
            newListAllBatches = []
            for listBatch in listAllBatches:
                for batch in listBatch:
                    newListAllBatches.append([batch])
            listAllBatches = newListAllBatches
        return listAllBatches



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "13/10/2023"

import os
import shlex
import subprocess
import threading


def run_command_line(command_line, timeout_sec=120):
    proc = subprocess.Popen(
        shlex.split(command_line), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    kill_proc = lambda p: p.kill()  # noqa E731
    timer = threading.Timer(timeout_sec, kill_proc, [proc])
    try:
        timer.start()
        binaryStdout, binaryStderr = proc.communicate()
        stdout = binaryStdout.decode("utf-8")
        stderr = binaryStderr.decode("utf-8")
    finally:
        timer.cancel()
    return stdout, stderr


def parse_salloc_stderr(stderr):
    job_id = None
    list_lines = stderr.split("\n")
    for line in list_lines:
        if line.startswith("salloc: Granted job allocation"):
            job_id = int(line.split(" ")[-1])
            break
    return job_id


def salloc(partition, exclusive=False):
    timeout_sec = 100
    salloc_command_line = f"salloc --no-shell -p {partition}"
    if exclusive:
        salloc_command_line += " --exclusive"
    stdout, stderr = run_command_line(salloc_command_line, timeout_sec)
    job_id = parse_salloc_stderr(stderr)
    if job_id is None:
        print(stdout)
        print(stderr)
    return job_id


def srun(job_id, command):
    timeout_sec = 10
    srun_command_line = f"srun --jobid {job_id} {command}"
    stdout, stderr = run_command_line(srun_command_line, timeout_sec)
    return stdout, stderr


def scancel(job_id):
    timeout_sec = 10
    scancel_command_line = f"scancel {job_id}"
    stdout, stderr = run_command_line(scancel_command_line, timeout_sec)


def submit_job_to_slurm(
    command_line,
    working_directory,
    nodes=1,
    core=4,
    time="2:00:00",
    host=None,
    queue=None,
    name=None,
    mem=None,
):
    slurm_id = None
    script = "#!/bin/bash\n"
    if name is not None:
        script += '#SBATCH --job-name="{0}"\n'.format(name)
    if queue is None:
        partition = "mx"
    else:
        partition = "mx,{0}".format(queue)
    script += "#SBATCH --partition={0}\n".format(partition)
    if mem is None:
        mem = 4000  # 4 Gb memory by default
    script += "#SBATCH --mem={0}\n".format(mem)
    script += "#SBATCH --ntasks={0}\n".format(nodes)
    script += "#SBATCH --nodes=1\n"  # Necessary for not splitting jobs! See ATF-57
    script += "#SBATCH --cpus-per-task={0}\n".format(core)
    script += "#SBATCH --time={0}\n".format(time)
    script += "#SBATCH --output={0}/stdout.txt\n".format(working_directory)
    script += "#SBATCH --error={0}/stderr.txt\n".format(working_directory)
    script += command_line + "\n"
    script_name = "slurm.sh"
    slurm_script_path = os.path.join(working_directory, script_name)
    with open(slurm_script_path, "w") as f:
        f.write(script)
    stdout, stderr = run_command_line("sbatch {0}".format(slurm_script_path))
    if "Submitted batch job" in stdout:
        slurm_id = int(stdout.split("job")[1])
    return slurm_script_path, slurm_id, stdout, stderr


def split_at_equal(line, dict_line={}):
    if "=" in line:
        key, value = line.split("=", 1)
        if "=" in value:
            if " " in value:
                part1, part2 = value.split(" ", 1)
            elif "," in value:
                part1, part2 = value.split(",", 1)
            dict_line[key.strip()] = part1.strip()
            dict_line = split_at_equal(part2, dict_line=dict_line)
        else:
            dict_line[key.strip()] = value.strip()
    return dict_line


def parse_slurm_stat(slurmStat):
    list_lines = slurmStat.split("\n")
    dict_slurm_stat = {}
    for line in list_lines:
        if line != "":
            dictLine = split_at_equal(line)
            dict_slurm_stat.update(dictLine)
    return dict_slurm_stat


def get_slurm_stat(slurm_job_id):
    dict_stat = None
    stdout, stderr = run_command_line("scontrol show job {0}".format(slurm_job_id))
    if stderr is None or stderr == "":
        dict_stat = parse_slurm_stat(stdout)
    return dict_stat


def are_jobs_pending(partition_name=None):
    jobs_are_pending = False
    if partition_name is not None:
        stdout, stderr = run_command_line(
            "squeue -p {0} -t PENDING".format(partition_name)
        )
    else:
        stdout, stderr = run_command_line("squeue -t PENDING")
    list_lines = stdout.split("\n")
    no_pending = 0
    for line in list_lines:
        if "Resources" in line or "Priority" in line:
            no_pending += 1
    if no_pending > 2:
        jobs_are_pending = True
    return jobs_are_pending



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "10/05/2019"

import json

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxv1/plugins/EDPluginControlImageQualityIndicators-v1.4/plugins/
#      EDPluginControlImageQualityIndicatorsv1_4.py

import os
import re
import shutil
import time
import pathlib

from pyicat_plus.client.main import IcatClient

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.WaitFileTask import WaitFileTask
from edna2.tasks.ControlDozor import ControlDozor
from edna2.tasks.PhenixTasks import DistlSignalStrengthTask
from edna2.tasks.H5ToCBFTask import H5ToCBFTask

from edna2.utils import UtilsPath
from edna2.utils import UtilsImage
from edna2.utils import UtilsIspyb
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

from edna2.tasks.ISPyBTasks import ISPyBRetrieveDataCollection

logger = UtilsLogging.getLogger()

DEFAULT_MIN_IMAGE_SIZE = 1000000
DEFAULT_WAIT_FILE_TIMEOUT = 300


class ImageQualityIndicators(AbstractTask):
    """
    This task controls the plugins that generate image quality indicators.
    """

    def __init__(self, inData, workingDirectorySuffix=None):
        AbstractTask.__init__(self, inData, workingDirectorySuffix)
        self.beamline = None
        self.directory = None
        self.template = None
        self.doSubmit = None
        self.doDozorM = None
        self.doDistlSignalStrength = None
        self.isFastMesh = None
        self.listImage = None
        self.batchSize = None
        self.minImageSize = None
        self.waitFileTimeout = None
        self.doIspybUpload = None
        self.dataCollectionId = None
        self.doIcatUpload = None
        self.overlap = None

    def run(self, inData):
        listImageQualityIndicators = []
        listControlDozorAllFile = []
        # Initialize parameters
        self.init(inData)
        # Set up batch list
        listOfBatches = self.createBatchList(inData)
        outData = dict()
        listDozorTask, distl_results = self.runDozorAndDistl(listOfBatches)
        if not self.isFailure():
            (
                listImageQualityIndicators,
                listControlDozorAllFile,
            ) = self.synchronizeDozor(listDozorTask, distl_results)
            # Assemble all controlDozorAllFiles into one
            if self.doDozorM:
                imageQualityIndicatorsDozorAllFile = self.createDozorAllFile(
                    listControlDozorAllFile
                )
                outData["dozorAllFile"] = imageQualityIndicatorsDozorAllFile
        outData["imageQualityIndicators"] = listImageQualityIndicators
        # Make plot if we have a data collection id
        if "dataCollectionId" in inData and inData["dataCollectionId"] is not None:
            working_directory = self.getWorkingDirectory()
            dozor_plot_path, dozor_csv_path = self.makePlot(
                inData["dataCollectionId"], outData, working_directory
            )
            logger.info("Checking if doIspybUpload")
            if self.doIspybUpload:
                logger.debug("doIspybUpload")
                self.storeDataOnPyarch(
                    inData["dataCollectionId"],
                    dozor_plot_path=dozor_plot_path,
                    dozor_csv_path=dozor_csv_path,
                    workingDirectory=working_directory,
                )

            if self.doIcatUpload:
                logger.debug("doIcatUpload")
                self.uploadDataToIcat(
                    working_directory=working_directory,
                    raw=str(self.directory),
                    dozor_csv_path=dozor_csv_path,
                    dozor_plot_path=dozor_plot_path,
                )

        return outData

    def init(self, inData):
        self.beamline = inData.get("beamline", None)
        self.doSubmit = inData.get("doSubmit", False)
        self.doDozorM = inData.get("doDozorM", False)
        self.doDistlSignalStrength = inData.get("doDistlSignalStrength", False)
        self.isFastMesh = inData.get("fastMesh", True)
        self.listImage = inData.get("image", [])
        self.batchSize = inData.get("batchSize", 100)
        self.doIspybUpload = inData.get("doIspybUpload", False)
        self.dataCollectionId = inData.get("dataCollectionId", None)
        self.doIcatUpload = inData.get("doIcatUpload", False)
        # Configurations
        self.minImageSize = UtilsConfig.get(
            self, "minImageSize", defaultValue=DEFAULT_MIN_IMAGE_SIZE
        )
        self.waitFileTimeOut = UtilsConfig.get(
            self, "waitFileTimeOut", defaultValue=DEFAULT_WAIT_FILE_TIMEOUT
        )
        if self.dataCollectionId is not None:
            ispybInData = {"dataCollectionId": self.dataCollectionId}
            ispybTask = ISPyBRetrieveDataCollection(inData=ispybInData)
            ispybTask.execute()
            dataCollection = ispybTask.outData
            inData["directory"] = dataCollection["imageDirectory"]
            inData["template"] = dataCollection["fileTemplate"].replace("%04d", "####")
            inData["startNo"] = dataCollection["startImageNumber"]
            inData["endNo"] = (
                dataCollection["startImageNumber"]
                + dataCollection["numberOfImages"]
                - 1
            )
            self.overlap = dataCollection["overlap"]
        else:
            self.overlap = inData.get("overlap", 0.0)

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "beamline": {"type": "string"},
                "doDozorM": {"type": "boolean"},
                "doDistlSignalStrength": {"type": "boolean"},
                "doIndexing": {"type": "boolean"},
                "doIspybUpload": {"type": "boolean"},
                "processDirectory": {"type": "string"},
                "image": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "batchSize": {"type": "integer"},
                "fastMesh": {"type": "boolean"},
                "wedgeNumber": {"type": "integer"},
                "directory": {"type": "string"},
                "template": {"type": "string"},
                "startNo": {"type": "integer"},
                "endNo": {"type": "integer"},
                "dataCollectionId": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
                "overlap": {"type": "number"}
            },
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "required": ["imageQualityIndicators"],
            "properties": {
                "imageQualityIndicators": {
                    "type": "array",
                    "items": {"$ref": self.getSchemaUrl("imageQualityIndicators.json")},
                },
                "inputDozor": {"type": "number"},
                "dozorAllFile": {"type": "string"},
            },
        }

    def createBatchList(self, inData):
        listOfBatches = []
        listOfImagesInBatch = []
        if len(self.listImage) == 0:
            self.directory = pathlib.Path(inData["directory"])
            self.template = inData["template"]
            startNo = inData["startNo"]
            endNo = inData["endNo"]
            for index in range(startNo, endNo + 1):
                listOfImagesInBatch.append(index)
                if len(listOfImagesInBatch) == self.batchSize or index == 9999:
                    listOfBatches.append(listOfImagesInBatch)
                    listOfImagesInBatch = []
        else:
            firstImage = pathlib.Path(self.listImage[0])
            self.directory = firstImage.parent
            self.template = UtilsImage.getTemplate(firstImage)
            for image in self.listImage:
                imageNo = UtilsImage.getImageNumber(image)
                listOfImagesInBatch.append(imageNo)
                if len(listOfImagesInBatch) == self.batchSize:
                    listOfBatches.append(listOfImagesInBatch)
                    listOfImagesInBatch = []
        if len(listOfImagesInBatch) > 0:
            listOfBatches.append(listOfImagesInBatch)
        return listOfBatches

    def runDozorAndDistl(self, listOfBatches):
        #
        # Loop over batches:
        # - Wait for all files in batch
        # - Run Dozor and DistlSignalStrength (if required) in parallel
        #
        dozor_tasks = []
        distl_results = {}
        template4d = re.sub("#+", "{0:04d}", self.template)
        template6d = re.sub("#+", "{0:06d}", self.template)
        for index, images_in_batch in enumerate(listOfBatches):
            distl_tasks = []
            listOfH5FilesInBatch = []
            image_no = images_in_batch[-1]
            # Wait for last image
            image_path = self.directory / template4d.format(image_no)
            logger.debug("Waiting for path: {0}".format(image_path))
            self.waitForImagePath(
                imagePath=image_path,
                batchSize=self.batchSize,
                isFastMesh=self.isFastMesh,
                minImageSize=self.minImageSize,
                waitFileTimeOut=self.waitFileTimeOut,
                listofH5FilesInBatch=listOfH5FilesInBatch,
            )
            logger.debug("Done waiting for path: {0}".format(image_path))
            if not self.isFailure():
                # Determine start and end image no
                batchStartNo = images_in_batch[0]
                batchEndNo = images_in_batch[-1]
                dozorTemplate = self.template
                # Convert images to CBF if necessary
                if self.doDistlSignalStrength and dozorTemplate.endswith(".h5"):
                    inDataH5ToCBF = {
                        "startImageNumber": batchStartNo,
                        "endImageNumber": batchEndNo,
                        "hdf5ImageNumber": 1,
                        "hdf5File": self.directory / template4d.format(batchStartNo),
                    }
                    h5ToCBFTask = H5ToCBFTask(inData=inDataH5ToCBF)
                    h5ToCBFTask.start()
                # Run Control Dozor
                inDataControlDozor = {
                    "template": dozorTemplate,
                    "directory": self.directory,
                    "startNo": batchStartNo,
                    "endNo": batchEndNo,
                    "batchSize": self.batchSize,
                    "doSubmit": self.doSubmit,
                    "doDozorM": self.doDozorM,
                    "doIspybUpload": self.doIspybUpload,
                    "overlap": self.overlap,
                }
                if self.beamline is not None:
                    inDataControlDozor["beamline"] = self.beamline
                controlDozor = ControlDozor(
                    inDataControlDozor,
                    workingDirectorySuffix="{0:04d}_{1:04d}".format(
                        batchStartNo, batchEndNo
                    ),
                )
                controlDozor.start()
                dozor_tasks.append(
                    (controlDozor, inDataControlDozor, list(images_in_batch))
                )
                # Check if we should run distl.signalStrength
                if self.doDistlSignalStrength:
                    if image_path.suffix == ".h5":
                        h5ToCBFTask.join()
                    for image_no in images_in_batch:
                        image_path = self.directory / template4d.format(image_no)
                        if image_path.suffix == ".h5":
                            image_path = str(
                                self.directory / template6d.format(image_no)
                            )
                            image_path = image_path.replace(".h5", ".cbf")
                        inDataDistl = {
                            "referenceImage": str(image_path),
                        }
                        distlTask = DistlSignalStrengthTask(
                            inData=inDataDistl,
                            workingDirectorySuffix=image_no,
                        )
                        # logger.info(f"Running {image_path}")
                        distlTask.start()
                        # logger.info(f"After running {image_path}")
                        distl_tasks.append((distlTask, image_path))
            for distlTask, image_path in distl_tasks:
                logger.info(f"Joining distlTask {image_path}")
                distlTask.join()
                distl_results[os.path.basename(image_path)] = dict(distlTask.outData)

        return dozor_tasks, distl_results

    def synchronizeDislt(self, listDistlTask):
        listDistlResult = []
        # Synchronize all image quality indicator plugins and upload to ISPyB
        for (image, distlTask) in listDistlTask:
            imageQualityIndicators = {}
            if distlTask is not None:
                # distlTask.join()
                if distlTask.isSuccess():
                    outDataDistl = distlTask.outData
                    if outDataDistl is not None:
                        imageQualityIndicators = outDataDistl["imageQualityIndicators"]
            imageQualityIndicators["image"] = str(image)
            listDistlResult.append(imageQualityIndicators)
        return listDistlResult

    def synchronizeDozor(self, listDozorTask, distl_results):
        listImageQualityIndicators = []
        listControlDozorAllFile = []
        for (
            controlDozor,
            inDataControlDozor,
            listBatch,
        ) in listDozorTask:
            controlDozor.join()
            # Check that we got at least one result
            if len(controlDozor.outData["imageQualityIndicators"]) == 0:
                # Run the dozor plugin again, this time synchronously
                firstImage = listBatch[0]
                lastImage = listBatch[-1]
                logger.warning(
                    "No dozor results! Re-executing Dozor for"
                    + " images {0} to {1}".format(firstImage, lastImage)
                )
                controlDozor = ControlDozor(
                    inDataControlDozor,
                    workingDirectorySuffix="{0:04d}_{1:04d}_redo".format(
                        firstImage, lastImage
                    ),
                )
                controlDozor.execute()
            listOutDataControlDozor = list(
                controlDozor.outData["imageQualityIndicators"]
            )
            if self.doDistlSignalStrength:
                for outDataControlDozor in listOutDataControlDozor:
                    image = os.path.basename(outDataControlDozor["image"])
                    image = image.replace(".h5", ".cbf")
                    if image in distl_results:
                        imageQualityIndicators = dict(outDataControlDozor)
                        imageQualityIndicators.update(distl_results[image]["imageQualityIndicators"])
                        listImageQualityIndicators.append(imageQualityIndicators)
                    else:
                        listImageQualityIndicators += listOutDataControlDozor
            else:
                listImageQualityIndicators += listOutDataControlDozor
            # Check if dozorm
            if self.doDozorM:
                listControlDozorAllFile.append(controlDozor.outData["dozorAllFile"])
        return listImageQualityIndicators, listControlDozorAllFile

    def createDozorAllFile(self, listControlDozorAllFile):
        imageQualityIndicatorsDozorAllFile = str(
            self.getWorkingDirectory() / "dozor_all"
        )
        os.system("touch {0}".format(imageQualityIndicatorsDozorAllFile))
        for controlDozorAllFile in listControlDozorAllFile:
            command = (
                "cat "
                + controlDozorAllFile
                + " >> "
                + imageQualityIndicatorsDozorAllFile
            )
            os.system(command)
        return imageQualityIndicatorsDozorAllFile

    @classmethod
    def getH5FilePath(cls, filePath, batchSize=1, isFastMesh=False):
        imageNumber = UtilsImage.getImageNumber(filePath)
        prefix = UtilsImage.getPrefix(filePath)
        if isFastMesh:
            h5ImageNumber = int((imageNumber - 1) / 100) + 1
            h5FileNumber = 1
        else:
            h5ImageNumber = 1
            h5FileNumber = int((imageNumber - 1) / batchSize) * batchSize + 1
        h5MasterFileName = "{prefix}_{h5FileNumber}_master.h5".format(
            prefix=prefix, h5FileNumber=h5FileNumber
        )
        h5MasterFilePath = filePath.parent / h5MasterFileName
        h5DataFileName = "{prefix}_{h5FileNumber}_data_{h5ImageNumber:06d}.h5".format(
            prefix=prefix, h5FileNumber=h5FileNumber, h5ImageNumber=h5ImageNumber
        )
        h5DataFilePath = filePath.parent / h5DataFileName
        return h5MasterFilePath, h5DataFilePath, h5FileNumber

    def waitForImagePath(
        self,
        imagePath,
        batchSize,
        isFastMesh,
        minImageSize,
        waitFileTimeOut,
        listofH5FilesInBatch,
    ):
        # Force an 'ls' in parent directory - this sometimes helps to 'unblock'
        # the file system
        os.system("ls {0} > /dev/null".format(os.path.dirname(imagePath)))
        # If Eiger, just wait for the h5 file
        if imagePath.suffix == ".h5":
            h5MasterFilePath, h5DataFilePath, hdf5ImageNumber = self.getH5FilePath(
                imagePath, batchSize=batchSize, isFastMesh=isFastMesh
            )
            if h5DataFilePath not in listofH5FilesInBatch:
                listofH5FilesInBatch.append(h5DataFilePath)
                logger.info("Eiger data, waiting for master" + " and data files...")
                inDataWaitFileTask = {
                    "file": str(h5DataFilePath),
                    "size": minImageSize,
                    "timeOut": waitFileTimeOut,
                }
                workingDirectorySuffix = h5DataFilePath.name.split(".h5")[0]
                waitFileTask = WaitFileTask(
                    inData=inDataWaitFileTask,
                    workingDirectorySuffix=workingDirectorySuffix,
                )
                logger.info("Waiting for file {0}".format(h5DataFilePath))
                logger.debug("Wait file timeOut set to %f" % waitFileTimeOut)
                waitFileTask.execute()
                time.sleep(0.1)
            if not os.path.exists(h5DataFilePath):
                errorMessage = "Time-out while waiting for image %s" % h5DataFilePath
                logger.error(errorMessage)
                self.setFailure()
        else:
            if not imagePath.exists():
                logger.info("Waiting for file {0}".format(imagePath))
                inDataWaitFileTask = {
                    "file": str(imagePath),
                    "size": minImageSize,
                    "timeOut": waitFileTimeOut,
                }
                workingDirectorySuffix = imagePath.name.split(imagePath.suffix)[0]
                waitFileTask = WaitFileTask(
                    inData=inDataWaitFileTask,
                    workingDirectorySuffix=workingDirectorySuffix,
                )
                logger.debug("Wait file timeOut set to %.0f s" % waitFileTimeOut)
                waitFileTask.execute()
            if not imagePath.exists():
                errorMessage = "Time-out while waiting for image " + str(imagePath)
                logger.error(errorMessage)
                self.setFailure()

    def createGnuPlotFile(self, workingDirectory, csvFileName, outDataImageDozor):
        with open(str(workingDirectory / csvFileName), "w") as gnuplotFile:
            gnuplotFile.write("# Data directory: {0}\n".format(self.directory))
            gnuplotFile.write(
                "# File template: {0}\n".format(self.template.replace("%04d", "####"))
            )
            gnuplotFile.write(
                "# {0:>9s}{1:>16s}{2:>16s}{3:>16s}{4:>16s}{5:>16s}\n".format(
                    "'Image no'",
                    "'Angle'",
                    "'No of spots'",
                    "'Main score (*10)'",
                    "'Spot score'",
                    "'Visible res.'",
                )
            )
            for imageQualityIndicators in outDataImageDozor["imageQualityIndicators"]:
                gnuplotFile.write(
                    "{0:10d},{1:15.3f},{2:15d},{3:15.3f},{4:15.3f},{5:15.3f}\n".format(
                        imageQualityIndicators["number"],
                        imageQualityIndicators["angle"],
                        imageQualityIndicators["dozorSpotsNumOf"],
                        10 * imageQualityIndicators["dozorScore"],
                        imageQualityIndicators["dozorSpotScore"],
                        imageQualityIndicators["dozorVisibleResolution"],
                    )
                )

    def determineMinMaxParameters(self, outDataImageDozor):
        minImageNumber = None
        maxImageNumber = None
        minAngle = None
        maxAngle = None
        minDozorValue = None
        maxDozorValue = None
        minResolution = None
        maxResolution = None
        for imageQualityIndicators in outDataImageDozor["imageQualityIndicators"]:
            if (
                minImageNumber is None
                or minImageNumber > imageQualityIndicators["number"]
            ):
                minImageNumber = imageQualityIndicators["number"]
                minAngle = imageQualityIndicators["angle"]
            if (
                maxImageNumber is None
                or maxImageNumber < imageQualityIndicators["number"]
            ):
                maxImageNumber = imageQualityIndicators["number"]
                maxAngle = imageQualityIndicators["angle"]
            if (
                minDozorValue is None
                or minDozorValue > imageQualityIndicators["dozorScore"]
            ):
                minDozorValue = imageQualityIndicators["dozorScore"]
            if (
                maxDozorValue is None
                or maxDozorValue < imageQualityIndicators["dozorScore"]
            ):
                maxDozorValue = imageQualityIndicators["dozorScore"]

            # Min resolution: the higher the value the lower the resolution
            if (
                minResolution is None
                or minResolution < imageQualityIndicators["dozorVisibleResolution"]
            ):
                # Disregard resolution worse than 10.0
                if imageQualityIndicators["dozorVisibleResolution"] < 10.0:
                    minResolution = imageQualityIndicators["dozorVisibleResolution"]
            # Max resolution: the lower the number the better the resolution
            if (
                maxResolution is None
                or maxResolution > imageQualityIndicators["dozorVisibleResolution"]
            ):
                maxResolution = imageQualityIndicators["dozorVisibleResolution"]
        plotDict = {
            "minImageNumber": minImageNumber,
            "maxImageNumber": maxImageNumber,
            "minAngle": minAngle,
            "maxAngle": maxAngle,
            "minDozorValue": minDozorValue,
            "maxDozorValue": maxDozorValue,
            "minResolution": minResolution,
            "maxResolution": maxResolution,
        }
        return plotDict

    def determinePlotParameters(self, plotDict):
        xtics = ""
        minImageNumber = plotDict["minImageNumber"]
        maxImageNumber = plotDict["maxImageNumber"]
        minAngle = plotDict["minAngle"]
        maxAngle = plotDict["maxAngle"]
        minResolution = plotDict["minResolution"]
        maxResolution = plotDict["maxResolution"]
        minDozorValue = plotDict["minDozorValue"]
        maxDozorValue = plotDict["maxDozorValue"]
        if minImageNumber is not None and minImageNumber == maxImageNumber:
            minAngle -= 1.0
            maxAngle += 1.0
        noImages = maxImageNumber - minImageNumber + 1
        if noImages <= 4:
            minImageNumber -= 0.1
            maxImageNumber += 0.1
            deltaAngle = maxAngle - minAngle
            minAngle -= deltaAngle * 0.1 / noImages
            maxAngle += deltaAngle * 0.1 / noImages
            xtics = "1"
        if maxResolution is None or maxResolution > 0.8:
            maxResolution = 0.8
        else:
            maxResolution = int(maxResolution * 10.0) / 10.0
        if minResolution is None or minResolution < 4.5:
            minResolution = 4.5
        else:
            minResolution = int(minResolution * 10.0) / 10.0 + 1
        if maxDozorValue < 0.001 and minDozorValue < 0.001:
            yscale = "set yrange [-0.5:0.5]\n    set ytics 1"
        else:
            yscale = "set autoscale  y"
        plotDict = {
            "xtics": xtics,
            "yscale": yscale,
            "minImageNumber": minImageNumber,
            "maxImageNumber": maxImageNumber,
            "minAngle": minAngle,
            "maxAngle": maxAngle,
            "minDozorValue": minDozorValue,
            "maxDozorValue": maxDozorValue,
            "minResolution": minResolution,
            "maxResolution": maxResolution,
        }
        return plotDict

    def makePlot(self, dataCollectionId, outDataImageDozor, working_directory):
        plotFileName = "dozor_{0}.png".format(dataCollectionId)
        csvFileName = "dozor_{0}.csv".format(dataCollectionId)
        self.createGnuPlotFile(working_directory, csvFileName, outDataImageDozor)
        plotDict = self.determineMinMaxParameters(outDataImageDozor)
        plotDict = self.determinePlotParameters(plotDict)
        gnuplotScript = """#
set terminal png
set output '{dozorPlotFileName}'
set title '{title}'
set grid x2 y2
set xlabel 'Image number'
set x2label 'Angle (degrees)'
set y2label 'Resolution (A)'
set ylabel 'Number of spots / ExecDozor score (*10)'
set xtics {xtics} nomirror
set x2tics
set ytics nomirror
set y2tics
set xrange [{minImageNumber}:{maxImageNumber}]
set x2range [{minAngle}:{maxAngle}]
{yscale}
set y2range [{minResolution}:{maxResolution}]
set key below
plot '{dozorCsvFileName}' using 1:3 title 'Number of spots' axes x1y1 with points linetype rgb 'goldenrod' pointtype 7 pointsize 1.5, \
    '{dozorCsvFileName}' using 1:4 title 'ExecDozor score' axes x1y1 with points linetype 3 pointtype 7 pointsize 1.5, \
    '{dozorCsvFileName}' using 1:6 title 'Visible resolution' axes x1y2 with points linetype 1 pointtype 7 pointsize 1.5
""".format(
            title=self.template.replace("%04d", "####"),
            dozorPlotFileName=plotFileName,
            dozorCsvFileName=csvFileName,
            minImageNumber=plotDict["minImageNumber"],
            maxImageNumber=plotDict["maxImageNumber"],
            minAngle=plotDict["minAngle"],
            maxAngle=plotDict["maxAngle"],
            minResolution=plotDict["minResolution"],
            maxResolution=plotDict["maxResolution"],
            xtics=plotDict["xtics"],
            yscale=plotDict["yscale"],
        )
        pathGnuplotScript = str(working_directory / "gnuplot.sh")
        with open(pathGnuplotScript, "w") as f:
            f.write(gnuplotScript)
        oldCwd = os.getcwd()
        os.chdir(str(working_directory))
        gnuplot = UtilsConfig.get(self, "gnuplot", "gnuplot")
        os.system("{0} {1}".format(gnuplot, pathGnuplotScript))
        os.chdir(oldCwd)
        dozor_plot_path = working_directory / plotFileName
        dozor_csv_path = working_directory / csvFileName
        return dozor_plot_path, dozor_csv_path

    @classmethod
    def storeDataOnPyarch(
        cls, dataCollectionId, dozor_plot_path, dozor_csv_path, workingDirectory
    ):
        resultsDirectory = pathlib.Path(workingDirectory) / "results"
        try:
            if not resultsDirectory.exists():
                resultsDirectory.mkdir(parents=True, mode=0o755)
            dozorPlotResultPath = resultsDirectory / dozor_plot_path.name
            dozorCsvResultPath = resultsDirectory / dozor_csv_path.name
            shutil.copy(dozor_plot_path, dozorPlotResultPath)
            shutil.copy(dozor_csv_path, dozorCsvResultPath)
        except Exception as e:
            logger.warning(
                "Couldn't copy files to results directory: {0}".format(resultsDirectory)
            )
            logger.warning(e)
        try:
            # Create paths on pyarch
            dozorPlotPyarchPath = UtilsPath.createPyarchFilePath(dozorPlotResultPath)
            dozorCsvPyarchPath = UtilsPath.createPyarchFilePath(dozorCsvResultPath)
            if not os.path.exists(os.path.dirname(dozorPlotPyarchPath)):
                os.makedirs(os.path.dirname(dozorPlotPyarchPath), 0o755)
            shutil.copy(dozorPlotResultPath, dozorPlotPyarchPath)
            shutil.copy(dozorCsvResultPath, dozorCsvPyarchPath)
            # Upload to data collection
            dataCollectionId = UtilsIspyb.setImageQualityIndicatorsPlot(
                dataCollectionId, dozorPlotPyarchPath, dozorCsvPyarchPath
            )
        except Exception as e:
            logger.warning("Couldn't copy files to pyarch.")
            logger.warning(e)

    def getBeamlineProposalFromPath(self, path):
        """ESRF specific code for extracting the beamline name and prefix from the path"""
        list_path = list(path.parts)
        if list_path[1] != "data":
            new_path = UtilsPath.stripDataDirectoryPrefix(str(path))
            list_path = pathlib.Path(new_path).parts
        beamline = None
        proposal = None
        if list_path[2] == "visitor":
            beamline = list_path[4]
            proposal = list_path[3]
        elif list_path[3] == "inhouse":
            beamline = list_path[2]
            proposal = list_path[4]
        return (beamline, proposal)

    def uploadDataToIcat(self, working_directory, raw, dozor_csv_path, dozor_plot_path):
        logger.debug(working_directory)
        if working_directory.parts[-2] == "nobackup":
            beamline, proposal = self.getBeamlineProposalFromPath(working_directory)
            logger.debug(beamline)
            logger.debug(proposal)
            if beamline is not None:
                dict_config = UtilsConfig.getTaskConfig("ICAT")
                metadata_urls = json.loads(dict_config.get("metadata_urls", "[]"))
                logger.debug(metadata_urls)
                if len(metadata_urls) > 0:
                    client = IcatClient(metadata_urls=metadata_urls)
                    icat_directory = working_directory.parents[1]
                    gallery_directory = icat_directory / "gallery"
                    gallery_directory.mkdir(mode=0o755, exist_ok=False)
                    shutil.copy(dozor_csv_path, gallery_directory)
                    shutil.copy(dozor_plot_path, gallery_directory)
                    sample_name = None
                    metadata_path = os.path.join(raw, "metadata.json")
                    if os.path.exists(metadata_path):
                        with open(metadata_path) as f:
                            metadata = json.loads(f.read())
                        if "Sample_name" in metadata:
                            sample_name = metadata["Sample_name"]
                    if sample_name is None:
                        dir1 = os.path.dirname(raw)
                        dir1_name = os.path.basename(dir1)
                        if dir1_name.startswith("run"):
                            dir2 = os.path.dirname(dir1)
                            dir2_name = os.path.basename(dir2)
                            if dir2_name.startswith("run"):
                                sample_name = os.path.basename(os.path.dirname(dir2))
                            else:
                                sample_name = dir2_name
                        else:
                            sample_name = dir1_name
                    data = {
                        "Sample_name": sample_name,
                        "scanType": "qualityIndicator",
                        "Process_program": "dozor"
                    }
                    dataset_name = "dozor_plot"
                    logger.debug(f"icat_directory {icat_directory}")
                    logger.debug("Before store")
                    icat_beamline = UtilsPath.getIcatBeamline(beamline)
                    logger.debug(icat_beamline)
                    client.store_processed_data(
                        beamline=icat_beamline,
                        proposal=proposal,
                        dataset=dataset_name,
                        path=str(icat_directory),
                        metadata=data,
                        raw=[raw],
                    )
                    logger.debug("After store")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "10/05/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginGroupCCP4-v1.0/plugins/EDPluginExecAimlessv1_0.py
# mxPluginExec/plugins/EDPluginGroupCCP4-v1.0/plugins/EDPluginExecPointlessv1_0.py

import os
import re

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class AimlessTask(AbstractTask):
    """
    Execution of CCP4 aimless
    """

    def run(self, inData):
        outData = {}
        input_file = inData['input_file']
        output_file = inData['output_file']
        symoplib = UtilsConfig.get('CCP4', 'symoplib')
        # ccp4setup = UtilsConfig.get('CCP4', 'ccp4setup')
        # if ccp4setup is None:
        #     commandLine = ""
        # else:
        #     commandLine = ". " + ccp4setup + '\n'
        commandLine = ""
        if symoplib is None:
            commandLine += 'aimless HKLIN {0} HKLOUT {1}'.format(
                input_file, output_file)
        else:
            commandLine += 'aimless HKLIN {0} HKLOUT {1} SYMINFO {2}'.format(
                input_file, output_file, symoplib)
        logger.info("Command line: {0}".format(commandLine))
        start_image = inData['start_image']
        end_image = inData['end_image']
        projectName = inData.get('dataCollectionID', 'EDNA_proc')
        resolution = inData.get('res', 0.0)
        anom = inData['anom']
        listCommand = [
            'bins 15',
            'run 1 batch {0} to {1}'.format(start_image, end_image),
            'name run 1 project {0} crystal DEFAULT dataset NATIVE'.format(projectName),
            'scales constant',
            'resolution 50 {0}'.format(resolution),
            'cycles 100',
            'anomalous {0}'.format('ON' if anom else 'OFF'),
            'output MERGED UNMERGED',
            'END'
            ]
        self.setLogFileName('aimless.log')
        self.runCommandLine(commandLine, listCommand=listCommand)
        outData['isSuccess'] = os.path.exists(output_file)
        return outData

class PointlessTask(AbstractTask):
    """
    Executes the CCP4 program pointless
    """

    def run(self, inData):
        input_file = inData['input_file']
        output_file = inData['output_file']
        commandLine = 'pointless'
        if UtilsConfig.isEMBL():
            commandLine += ' -c'
        commandLine += " xdsin {0} hklout {1}".format(input_file, output_file)
        listCommand = ['setting symmetry-based']
        if 'choose_spacegroup' in inData:
            listCommand += 'choose spacegroup {0}'.format(
                inData['choose_spacegroup'])
        self.setLogFileName('pointless.log')
        self.runCommandLine(commandLine, listCommand=listCommand)
        outData = self.parsePointlessOutput(self.getLogPath())
        return outData

    @classmethod
    def parsePointlessOutput(cls, logPath):
        sgre = re.compile(""" \* Space group = '(?P<sgstr>.*)' \(number\s+(?P<sgnumber>\d+)\)""")
        outData = {'isSuccess': False}
        if logPath.exists():
            with open(str(logPath)) as f:
                log = f.read()
            m = sgre.search(log)
            if m is not None:
                d = m.groupdict()
                sgnumber = d['sgnumber']
                sgstr = d['sgstr']

                outData['sgnumber'] = int(sgnumber)
                outData['sgstr'] = sgstr
                outData['isSuccess'] = True
                # Search first for unit cell after the Laue group...
                unitCellRe = re.compile("""  Laue group confidence.+\\n\\n\s+Unit cell:(.+)""")
                m2 = unitCellRe.search(log)
                if m2 is None:
                    # Then search it from the end...
                    unitCellRe = re.compile(""" \* Cell Dimensions : \(obsolete \- refer to dataset cell dimensions above\)\\n\\n(.+)""")
                    m2 = unitCellRe.search(log)
                if m2 is not None:
                    listCell = m2.groups()[0].split()
                    cell = {
                        'length_a': float(listCell[0]),
                        'length_b': float(listCell[1]),
                        'length_c': float(listCell[2]),
                        'angle_alpha': float(listCell[3]),
                        'angle_beta': float(listCell[4]),
                        'angle_gamma': float(listCell[5])
                    }
                    outData['cell'] = cell
        return outData




#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "23/11/2020"

import math

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.Best import Best
from edna2.tasks.ControlIndexing import ControlIndexing
from edna2.tasks.XDSTasks import XDSGenerateBackground
from edna2.tasks.XDSTasks import XDSIntegration
from edna2.tasks.ReadImageHeader import ReadImageHeader
from edna2.tasks.Raddose import Raddose
from edna2.tasks.H5ToCBFTask import H5ToCBFTask

from edna2.utils import UtilsImage
from edna2.utils import UtilsSymmetry


class Characterisation(AbstractTask):
    """
    This task receives a list of images or data collection ids and
    returns result of indexing
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "dataCollectionId": {"type": "integer"},
                "diffractionPlan": {
                    "$ref": self.getSchemaUrl("ispybDiffractionPlan.json")
                },
                "experimentalCondition": {
                    "$ref": self.getSchemaUrl("experimentalCondition.json")
                },
                "imagePath": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
                "sample": {"$ref": self.getSchemaUrl("sample.json")},
                "token": {"anyOf": [{"type": "string"}, {"type": "null"}]},
            },
        }

    def run(self, inData):
        listImagePath = inData["imagePath"]
        prefix = UtilsImage.getPrefix(listImagePath[0])
        listSubWedge = self.getListSubWedge(inData)
        diffractionPlan = inData.get("diffractionPlan", {})
        sample = inData.get("sample", {})
        # Check if flux is present
        flux = None
        absorbedDoseRate = None
        experimentalCondition = inData.get("experimentalCondition", None)
        if experimentalCondition is not None:
            beam = experimentalCondition.get("beam", None)
            if beam is not None:
                flux = beam.get("flux", None)
                # if not "size" in beam:
                #     beam["size"] = listSubWedge[0]["experimentalCondition"]["beam"]["size"]
                beam["exposureTime"] = listSubWedge[0]["experimentalCondition"]["beam"][
                    "exposureTime"
                ]
                beam["wavelength"] = listSubWedge[0]["experimentalCondition"]["beam"][
                    "wavelength"
                ]
        # Convert images to CBF
        firstImage = listSubWedge[0]["image"][0]["path"]
        import os
        import pprint

        listH5ToCBF = []
        suffix = os.path.splitext(firstImage)[1]
        if suffix == ".h5":
            for subWedge in listSubWedge:
                imageList = subWedge["image"]
                for image in imageList:
                    imagePath = image["path"]
                    hdf5ImageNumber = UtilsImage.getImageNumber(imagePath)
                    inDataH5ToCBF = {
                        "hdf5File": imagePath,
                        "hdf5ImageNumber": hdf5ImageNumber,
                        "imageNumber": 1,
                        "forcedOutputDirectory": str(self.getWorkingDirectory()),
                        "forcedOutputImageNumber": hdf5ImageNumber,
                    }
                    h5ToCBF = H5ToCBFTask(inData=inDataH5ToCBF)
                    h5ToCBF.start()
                    listH5ToCBF.append((image, h5ToCBF))
        # Join CBF creation
        for image, h5ToCBF in listH5ToCBF:
            h5ToCBF.join()
            image["path"] = h5ToCBF.outData["outputCBFFile"]
        pprint.pprint(listSubWedge)
        # Start indexing
        outDataIndexing, outDataGB, listSubWedge = self.indexing(
            prefix, listSubWedge, self.getWorkingDirectory()
        )
        if outDataIndexing is not None and outDataGB is not None:
            listXdsAsciiHkl, correctLp, bkgpixCbf = self.integration(
                prefix, listSubWedge, outDataIndexing, outDataGB
            )
            if listXdsAsciiHkl is not None:
                # Check if Raddose should be run
                estimateRadiationDamage = self.checkEstimateRadiationDamage(
                    inData, flux
                )
                if estimateRadiationDamage:
                    # Check if forced space group
                    forcedSpaceGroup = None
                    numOperators = None
                    cell = None
                    if (
                        "diffractionPlan" in inData
                        and "forcedSpaceGroup" in inData["diffractionPlan"]
                    ):
                        forcedSpaceGroup = inData["diffractionPlan"]["forcedSpaceGroup"]
                        if forcedSpaceGroup is not None:
                            if forcedSpaceGroup != "":
                                forcedSpaceGroup = forcedSpaceGroup.replace(" ", "")
                                numOperators = UtilsSymmetry.getNumberOfSymmetryOperatorsFromSpaceGroupName(
                                    forcedSpaceGroup
                                )
                            else:
                                forcedSpaceGroup = None
                    if forcedSpaceGroup is None:
                        # Get indexing space group IT number
                        if "resultIndexing" in outDataIndexing:
                            resultIndexing = outDataIndexing["resultIndexing"]
                            cell = resultIndexing["cell"]
                            if "spaceGroupNumber" in resultIndexing:
                                spaceGroupNumber = resultIndexing["spaceGroupNumber"]
                                numOperators = UtilsSymmetry.getNumberOfSymmetryOperatorsFromSpaceGroupITNumber(
                                    spaceGroupNumber
                                )
                    if numOperators is None:
                        raise RuntimeError(
                            "Error when trying to determine number of symmetry operators!"
                        )
                    chemicalComposition = self.getDefaultChemicalComposition(
                        cell, numOperators
                    )
                    numberOfImages = self.getNumberOfImages(listSubWedge)
                    sample = inData["sample"]
                    inDataRaddose = {
                        "experimentalCondition": experimentalCondition,
                        "chemicalComposition": chemicalComposition,
                        "sample": sample,
                        "cell": cell,
                        "numberOfImages": numberOfImages,
                        "numOperators": numOperators,
                    }
                    # import pprint
                    # pprint.pprint(inDataRaddose)
                    raddose = Raddose(
                        inData=inDataRaddose, workingDirectorySuffix=prefix
                    )
                    raddose.execute()
                    if raddose.isSuccess():
                        absorbedDoseRate = raddose.outData["absorbedDoseRate"]
                inDataBest = {
                    "diffractionPlan": diffractionPlan,
                    "sample": sample,
                    "subWedge": listSubWedge,
                    "xdsAsciiHkl": listXdsAsciiHkl,
                    "bkgpixCbf": bkgpixCbf,
                    "correctLp": correctLp,
                    "crystalAbsorbedDoseRate": absorbedDoseRate,
                }
                bestTask = Best(inData=inDataBest, workingDirectorySuffix=prefix)
                bestTask.execute()

    @staticmethod
    def indexing(prefix, listSubWedge, workingDirectory):
        outDataIndexing = None
        outDataGB = None
        # Start indexing
        inDataIndexing = {
            "subWedge": listSubWedge,
        }
        indexingTask = ControlIndexing(
            inData=inDataIndexing, workingDirectorySuffix=prefix
        )
        indexingTask.start()
        # Start background esitmation
        inDataGenerateBackground = {
            "subWedge": [listSubWedge[0]],
        }
        generateBackground = XDSGenerateBackground(
            inData=inDataGenerateBackground, workingDirectorySuffix=prefix
        )
        generateBackground.start()
        generateBackground.join()
        # Check indexing
        indexingTask.join()
        if indexingTask.isSuccess():
            outDataIndexing = indexingTask.outData
            outDataGB = generateBackground.outData
        return outDataIndexing, outDataGB, listSubWedge

    @staticmethod
    def integration(prefix, listSubWedge, outDataIndexing, outDataGB):
        listXdsAsciiHkl = None
        correctLp = None
        bkgpixCbf = None
        resultIndexing = outDataIndexing["resultIndexing"]
        xparmXdsPath = outDataIndexing["xparmXdsPath"]
        if xparmXdsPath is not None:
            listTasks = []
            listXdsAsciiHkl = []
            for subWedge in listSubWedge:
                inDataIntergation = {
                    "subWedge": [subWedge],
                    "xparmXds": xparmXdsPath,
                    "spaceGroupNumber": resultIndexing["spaceGroupNumber"],
                    "cell": resultIndexing["cell"],
                    "gainCbf": outDataGB["gainCbf"],
                    "blankCbf": outDataGB["blankCbf"],
                    "bkginitCbf": outDataGB["bkginitCbf"],
                    "xCorrectionsCbf": outDataGB["xCorrectionsCbf"],
                    "yCorrectionsCbf": outDataGB["yCorrectionsCbf"],
                }
                imageNo = subWedge["image"][0]["number"]
                integrationTask = XDSIntegration(
                    inData=inDataIntergation,
                    workingDirectorySuffix=prefix + "_{0:04d}".format(imageNo),
                )
                integrationTask.start()
                listTasks.append(integrationTask)
            for task in listTasks:
                task.join()
                if "xdsAsciiHkl" in task.outData:
                    listXdsAsciiHkl.append(task.outData["xdsAsciiHkl"])
                if correctLp is None:
                    correctLp = task.outData["correctLp"]
                    bkgpixCbf = task.outData["bkgpixCbf"]
        return listXdsAsciiHkl, correctLp, bkgpixCbf

    @staticmethod
    def getListSubWedge(inData):
        listSubWedge = None
        # First check if we have data collection ids or image list
        # if "dataCollectionId" in inData:
        #     # TODO: get list of data collections from ISPyB
        #         logger.warning("Not implemented!")
        # el
        if "imagePath" in inData:
            listSubWedge = Characterisation.readImageHeaders(inData["imagePath"])
        else:
            raise RuntimeError("No dataCollectionId or imagePath in inData")
        return listSubWedge

    @staticmethod
    def readImageHeaders(listImagePath):
        # Read the header(s)
        inDataReadImageHeader = {"imagePath": listImagePath}
        readImageHeader = ReadImageHeader(
            inData=inDataReadImageHeader,
            workingDirectorySuffix=UtilsImage.getPrefix(listImagePath[0]),
        )
        readImageHeader.execute()
        listSubWedge = readImageHeader.outData["subWedge"]
        return listSubWedge

    @staticmethod
    def getDefaultChemicalComposition(cell, numOperators):

        # For default chemical composition
        averageAminoAcidVolume = 135.49
        averageCrystalSolventContent = 0.47
        averageSulfurContentPerAminoacid = 0.05
        averageSulfurConcentration = 314

        a = cell["a"]
        b = cell["b"]
        c = cell["c"]
        alpha = math.radians(cell["alpha"])
        beta = math.radians(cell["beta"])
        gamma = math.radians(cell["gamma"])

        unitCellVolume = (
            a
            * b
            * c
            * (
                math.sqrt(
                    1
                    - math.cos(alpha) * math.cos(alpha)
                    - math.cos(beta) * math.cos(beta)
                    - math.cos(gamma) * math.cos(gamma)
                    + 2 * math.cos(alpha) * math.cos(beta) * math.cos(gamma)
                )
            )
        )
        polymerVolume = unitCellVolume * (1 - averageCrystalSolventContent)
        numberOfMonomersPerUnitCell = round(polymerVolume / averageAminoAcidVolume)
        numberOfMonomersPerAsymmetricUnit = round(
            numberOfMonomersPerUnitCell / numOperators
        )
        numberOfSulfurAtom = int(
            round(numberOfMonomersPerAsymmetricUnit * averageSulfurContentPerAminoacid)
        )

        chemicalCompositionMM = {
            "solvent": {
                "atom": [
                    {
                        "symbol": "S",
                        "concentration": averageSulfurConcentration,
                    }
                ]
            },
            "structure": {
                "chain": [
                    {
                        "type": "protein",
                        "numberOfCopies": 1,
                        "numberOfMonomers": numberOfMonomersPerAsymmetricUnit,
                        "heavyAtoms": [{"symbol": "S", "numberOf": numberOfSulfurAtom}],
                    }
                ],
                "numberOfCopiesInAsymmetricUnit": 1,
            },
        }

        return chemicalCompositionMM

    @staticmethod
    def checkEstimateRadiationDamage(inData, flux=None):
        estimateRadiationDamage = None
        # Check if radiation damage estimation is required or not in the diffraction plan
        diffractionPlan = inData.get("diffractionPlan", None)
        if diffractionPlan is not None:
            if "estimateRadiationDamage" in diffractionPlan:
                # Estimate radiation damage is explicitly set in the diffraction plan
                estimateRadiationDamage = diffractionPlan["estimateRadiationDamage"]
            else:
                strategyOption = diffractionPlan.get("strategyOption", None)
                if strategyOption is not None and "-DamPar" in strategyOption:
                    # The "-DamPar" option requires estimation of radiation damage
                    estimateRadiationDamage = True

        # Check if not set by the diffraction plan
        if estimateRadiationDamage is None:
            experimentalCondition = inData.get("experimentalCondition", None)
            if experimentalCondition is not None:
                beam = experimentalCondition.get("beam", None)
                if beam is not None:
                    flux = beam.get("flux")
                    if flux is not None:
                        estimateRadiationDamage = True

        if estimateRadiationDamage is None:
            estimateRadiationDamage = False

        return estimateRadiationDamage

    @staticmethod
    def getNumberOfImages(listSubwedge):
        noImages = 0
        for subWedge in listSubwedge:
            goniostat = subWedge["experimentalCondition"]["goniostat"]
            oscStart = goniostat["rotationAxisStart"]
            oscEnd = goniostat["rotationAxisEnd"]
            oscWidth = goniostat["oscillationWidth"]
            noImages += int(round((oscEnd - oscStart) / oscWidth, 0))
        return noImages



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# kernel/tests/src/EDTestCasePlugin.py (for loadTestImage)

import os
import re
import json
import pathlib
import datetime
import tempfile
import threading

from edna2.utils import UtilsLogging
from edna2.utils import UtilsImage

from urllib.request import urlopen, ProxyHandler, build_opener

logger = UtilsLogging.getLogger()

URL_EDNA_SITE = "http://www.edna-site.org/data/tests/images"
MAX_DOWNLOAD_TIME = 300


def __timeoutDuringDownload():
    """
    Function called after a timeout in the download part, raises RuntimeError.
    """
    raise RuntimeError(
        "Could not automatically download test images!\n"
        + "If you are behind a firewall,"
        + "please set the environment variable http_proxy.\n"
        + "Otherwise please try to download the images manually from\n"
        + "http://www.edna-site.org/data/tests/images"
    )


def getTestdataPath():
    pathFile = pathlib.Path(__file__)
    pathProjectBase = pathFile.parents[3]
    testdataPath = pathProjectBase / "tests" / "testdata"
    return testdataPath


def getTestRunPath():
    now = datetime.datetime.now()
    dateTime = now.strftime("%Y%m%d_%H%M%S")
    runPath = getTestdataPath() / "rundir" / dateTime
    if not runPath.exists():
        runPath.mkdir(0o755, parents=True)
    return runPath


def getTestImageDirPath():
    dirPath = getTestdataPath() / "images"
    if not dirPath.exists():
        dirPath.mkdir(0o755, parents=True)
    return dirPath


def loadTestImage(imageFileName):
    """
    This method tries to download images from
    http://www.edna-site.org/data/tests/images
    """
    imageDirPath = getTestImageDirPath()
    if not imageDirPath.exists():
        imageDirPath.mkdir(mode=0o777, parents=True)
    # Check if h5 data
    if imageFileName.endswith(".h5"):
        imagePath = imageDirPath / imageFileName
        hasOverlap = imagePath.name.startswith("ref-")
        h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(
            imagePath, hasOverlap=hasOverlap
        )
        listImagePath = [h5MasterFilePath, h5DataFilePath]
    else:
        listImagePath = [imageDirPath / imageFileName]
    for imagePath in listImagePath:
        if not imagePath.exists():
            logger.info(
                "Trying to download image %s" % str(imagePath)
                + ", timeout set to %d s" % MAX_DOWNLOAD_TIME
            )
            if "http_proxy" in os.environ:
                dictProxies = {"http": os.environ["http_proxy"]}
                proxy_handler = ProxyHandler(dictProxies)
                opener = build_opener(proxy_handler).open
            else:
                opener = urlopen

            timer = threading.Timer(MAX_DOWNLOAD_TIME + 1, __timeoutDuringDownload)
            timer.start()
            data = opener(
                "%s/%s" % (URL_EDNA_SITE, imagePath.name),
                data=None,
                timeout=MAX_DOWNLOAD_TIME,
            ).read()
            timer.cancel()

            try:
                open(str(imagePath), "wb").write(data)
            except IOError:
                raise IOError(
                    "Unable to write downloaded data to disk at %s" % imagePath
                )

        if os.path.exists(str(imagePath)):
            logger.info("Image %s successfully downloaded." % imagePath)
        else:
            raise RuntimeError(
                "Could not automatically download test image %r!\n"
                + "If you are behind a firewall, "
                + "please set the environment variable http_proxy.\n"
                + "Otherwise please try to download the images manually from\n"
                + "http://www.edna-site.org/data/tests/images" % imageFileName
            )


def substitute(data, searchString, substituteString):
    dataString = json.dumps(data)
    newDataString = dataString.replace(searchString, substituteString)
    newData = json.loads(newDataString)
    return newData


def getSearchStringFileNames(searchString, data):
    listFileNames = []
    dataString = json.dumps(data)
    # With suffix
    expression = r'"\{0}/([\w-]+\.\w+)"'.format(searchString)
    for match in re.findall(expression, dataString):
        listFileNames.append(match)
    if len(listFileNames) == 0:
        # Try without suffix
        expression = r'"\{0}/([\w-]+)"'.format(searchString)
        for match in re.findall(expression, dataString):
            listFileNames.append(match)
    return listFileNames


def substitueTestData(inData, loadTestImages=True, taskDataPath=None, tmpDir=None):
    # $EDNA2_TESTDATA_IMAGES
    searchString = "$EDNA2_TESTDATA_IMAGES"
    substituteString = getTestImageDirPath().as_posix()
    listFileNames = getSearchStringFileNames(searchString, inData)
    if loadTestImages:
        for imageFileName in listFileNames:
            loadTestImage(imageFileName)
    newInData = substitute(inData, searchString, substituteString)
    # $EDNA2_TASK_DATA
    if taskDataPath is not None:
        searchString = "$EDNA2_TASK_DATA"
        substituteString = str(taskDataPath)
        newInData = substitute(newInData, searchString, substituteString)
    # $EDNA2_TMP_DATA
    if tmpDir is not None:
        searchString = "$EDNA2_TMP_DATA"
        substituteString = str(tmpDir)
        newInData = substitute(newInData, searchString, substituteString)
    # Any other environment variables...
    newInData = json.loads(os.path.expandvars(json.dumps(newInData)))
    return newInData


def loadAndSubstitueTestData(dataPath, loadTestImages=True, tmpDir=None):
    with open(str(dataPath)) as f:
        inData = json.loads(f.read())
    taskDataPath = dataPath.parent
    return substitueTestData(
        inData, loadTestImages=loadTestImages, taskDataPath=taskDataPath, tmpDir=tmpDir
    )


def createTestTmpDirectory(testName):
    return tempfile.mkdtemp(prefix=testName + "_")


def prepareTestDataPath(modulePath):
    dataPath = pathlib.Path(modulePath).parent / "data"
    os.chdir(str(getTestRunPath()))
    return dataPath
